#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#               2009       Gary Burton
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
The EditLdsOrd module provides the EditLdsOrd class. This provides a
mechanism for the user to edit personal LDS information.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.lib import LdsOrd, NoteType
from gramps.gen.display.name import displayer as name_displayer
from gramps.gen.utils.lds import TEMPLES
from ..glade import Glade
from .editsecondary import EditSecondary
from .objectentries import PlaceEntry
from .displaytabs import CitationEmbedList, NoteTab
from ..widgets import PrivacyButton, MonitoredDate, MonitoredMenu, MonitoredStrMenu
from ..selectors import SelectorFactory
from gramps.gen.const import URL_MANUAL_SECT1

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

WIKI_HELP_PAGE = URL_MANUAL_SECT1
WIKI_HELP_SEC = _("LDS_Ordinance_Editor", "manual")

_DATA_MAP = {
    LdsOrd.BAPTISM: [
        LdsOrd.STATUS_NONE,
        LdsOrd.STATUS_CHILD,
        LdsOrd.STATUS_CLEARED,
        LdsOrd.STATUS_COMPLETED,
        LdsOrd.STATUS_INFANT,
        LdsOrd.STATUS_PRE_1970,
        LdsOrd.STATUS_QUALIFIED,
        LdsOrd.STATUS_STILLBORN,
        LdsOrd.STATUS_SUBMITTED,
        LdsOrd.STATUS_UNCLEARED,
    ],
    LdsOrd.CONFIRMATION: [
        LdsOrd.STATUS_NONE,
        LdsOrd.STATUS_CHILD,
        LdsOrd.STATUS_CLEARED,
        LdsOrd.STATUS_COMPLETED,
        LdsOrd.STATUS_INFANT,
        LdsOrd.STATUS_PRE_1970,
        LdsOrd.STATUS_QUALIFIED,
        LdsOrd.STATUS_STILLBORN,
        LdsOrd.STATUS_SUBMITTED,
        LdsOrd.STATUS_UNCLEARED,
    ],
    LdsOrd.ENDOWMENT: [
        LdsOrd.STATUS_NONE,
        LdsOrd.STATUS_CHILD,
        LdsOrd.STATUS_CLEARED,
        LdsOrd.STATUS_COMPLETED,
        LdsOrd.STATUS_INFANT,
        LdsOrd.STATUS_PRE_1970,
        LdsOrd.STATUS_QUALIFIED,
        LdsOrd.STATUS_STILLBORN,
        LdsOrd.STATUS_SUBMITTED,
        LdsOrd.STATUS_UNCLEARED,
    ],
    LdsOrd.SEAL_TO_PARENTS: [
        LdsOrd.STATUS_NONE,
        LdsOrd.STATUS_BIC,
        LdsOrd.STATUS_CLEARED,
        LdsOrd.STATUS_COMPLETED,
        LdsOrd.STATUS_DNS,
        LdsOrd.STATUS_PRE_1970,
        LdsOrd.STATUS_QUALIFIED,
        LdsOrd.STATUS_STILLBORN,
        LdsOrd.STATUS_SUBMITTED,
        LdsOrd.STATUS_UNCLEARED,
    ],
    LdsOrd.SEAL_TO_SPOUSE: [
        LdsOrd.STATUS_NONE,
        LdsOrd.STATUS_CANCELED,
        LdsOrd.STATUS_CLEARED,
        LdsOrd.STATUS_COMPLETED,
        LdsOrd.STATUS_DNS,
        LdsOrd.STATUS_PRE_1970,
        LdsOrd.STATUS_QUALIFIED,
        LdsOrd.STATUS_DNS_CAN,
        LdsOrd.STATUS_SUBMITTED,
        LdsOrd.STATUS_UNCLEARED,
    ],
}


# -------------------------------------------------------------------------
#
# EditLdsOrd class
#
# -------------------------------------------------------------------------
class EditLdsOrd(EditSecondary):
    """
    Displays a dialog that allows the user to edit an attribute.
    """

    def __init__(self, state, uistate, track, attrib, callback):
        """
        Displays the dialog box.

        parent - The class that called the Address editor.
        attrib - The attribute that is to be edited
        title - The title of the dialog box
        list - list of options for the pop down menu
        """
        EditSecondary.__init__(self, state, uistate, track, attrib, callback)

    def _local_init(self):
        self.top = Glade()
        self.set_window(
            self.top.toplevel, self.top.get_object("title"), _("LDS Ordinance Editor")
        )
        self.setup_configs("interface.lds", 600, 450)

        self.share_btn = self.top.get_object("share_place")
        self.add_del_btn = self.top.get_object("add_del_place")

    def _connect_signals(self):
        self.parents_select.connect("clicked", self.select_parents_clicked)
        self.define_cancel_button(self.top.get_object("cancel"))
        self.define_ok_button(self.top.get_object("ok"), self.save)
        self.define_help_button(
            self.top.get_object("help"), WIKI_HELP_PAGE, WIKI_HELP_SEC
        )

    def _get_types(self):
        return (
            LdsOrd.BAPTISM,
            LdsOrd.ENDOWMENT,
            LdsOrd.CONFIRMATION,
            LdsOrd.SEAL_TO_PARENTS,
        )

    def _setup_fields(self):
        self.parents_label = self.top.get_object("parents_label")
        self.parents = self.top.get_object("parents")
        self.parents_select = self.top.get_object("parents_select")

        self.priv = PrivacyButton(
            self.top.get_object("private"), self.obj, self.db.readonly
        )

        self.date_field = MonitoredDate(
            self.top.get_object("date_entry"),
            self.top.get_object("date_stat"),
            self.obj.get_date_object(),
            self.uistate,
            self.track,
            self.db.readonly,
        )

        self.place_field = PlaceEntry(
            self.dbstate,
            self.uistate,
            self.track,
            self.top.get_object("place"),
            self.top.get_object("place_event_box"),
            self.obj.set_place_handle,
            self.obj.get_place_handle,
            self.add_del_btn,
            self.share_btn,
        )

        self.type_menu = MonitoredMenu(
            self.top.get_object("type"),
            self.obj.set_type,
            self.obj.get_type,
            [
                (item[1], item[0])
                for item in LdsOrd._TYPE_MAP
                if item[0] in self._get_types()
            ],
            self.db.readonly,
            changed=self.ord_type_changed,
        )
        self.track_ref_for_deletion("type_menu")

        self.temple_menu = MonitoredStrMenu(
            self.top.get_object("temple"),
            self.obj.set_temple,
            self.obj.get_temple,
            TEMPLES.name_code_data(),
            self.db.readonly,
        )
        self.track_ref_for_deletion("temple_menu")

        self.status_menu = MonitoredMenu(
            self.top.get_object("status"),
            self.obj.set_status,
            self.obj.get_status,
            [
                (item[1], item[0])
                for item in LdsOrd._STATUS_MAP
                if item[0] in _DATA_MAP[self.obj.get_type()]
            ],
            self.db.readonly,
        )
        self.track_ref_for_deletion("status_menu")

        self.update_parent_label()

    def _post_init(self):
        self.ord_type_changed()

    def ord_type_changed(self):
        if self.obj.get_type() == LdsOrd.SEAL_TO_PARENTS:
            self.parents.show()
            self.parents_label.show()
            self.parents_select.show()
        else:
            self.parents.hide()
            self.parents_label.hide()
            self.parents_select.hide()

        new_data = [
            (item[1], item[0])
            for item in LdsOrd._STATUS_MAP
            if item[0] in _DATA_MAP[self.obj.get_type()]
        ]
        self.status_menu.change_menu(new_data)

    def _create_tabbed_pages(self):
        notebook = Gtk.Notebook()
        self.citation_list = CitationEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_citation_list(),
            "ldsord_editor_citations",
        )
        self._add_tab(notebook, self.citation_list)
        self.track_ref_for_deletion("citation_list")

        self.note_tab = NoteTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_note_list(),
            "ldsord_editor_notes",
            notetype=NoteType.LDS,
        )
        self._add_tab(notebook, self.note_tab)
        self.track_ref_for_deletion("note_tab")

        self._setup_notebook_tabs(notebook)
        notebook.show_all()
        self.top.get_object("vbox").pack_start(notebook, True, True, 0)

    def select_parents_clicked(self, obj):
        SelectFamily = SelectorFactory("Family")

        dialog = SelectFamily(self.dbstate, self.uistate, self.track)
        family = dialog.run()
        if family:
            self.obj.set_family_handle(family.handle)
        self.update_parent_label()

    def update_parent_label(self):
        handle = self.obj.get_family_handle()
        if handle:
            family = self.dbstate.db.get_family_from_handle(handle)
            f = self.dbstate.db.get_person_from_handle(family.get_father_handle())
            m = self.dbstate.db.get_person_from_handle(family.get_mother_handle())
            if f and m:
                label = _("%(father)s and %(mother)s [%(gramps_id)s]") % {
                    "father": name_displayer.display(f),
                    "mother": name_displayer.display(m),
                    "gramps_id": family.gramps_id,
                }
            elif f:
                label = _("%(father)s [%(gramps_id)s]") % {
                    "father": name_displayer.display(f),
                    "gramps_id": family.gramps_id,
                }
            elif m:
                label = _("%(mother)s [%(gramps_id)s]") % {
                    "mother": name_displayer.display(m),
                    "gramps_id": family.gramps_id,
                }
            else:
                # No translation for bare gramps_id
                label = "[%(gramps_id)s]" % {
                    "gramps_id": family.gramps_id,
                }
        else:
            label = ""

        self.parents.set_text(label)

    def build_menu_names(self, attrib):
        label = _("LDS Ordinance")
        return (label, _("LDS Ordinance Editor"))

    def save(self, *obj):
        """
        Called when the OK button is pressed. Gets data from the
        form and updates the LdsOrd data structure.
        """
        if self.callback:
            self.callback(self.obj)
        self.callback = None
        self.close()


# -------------------------------------------------------------------------
#
# EditFamilyLdsOrd
#
# -------------------------------------------------------------------------
class EditFamilyLdsOrd(EditSecondary):
    """
    Displays a dialog that allows the user to edit an attribute.
    """

    def __init__(self, state, uistate, track, attrib, callback):
        """
        Displays the dialog box.

        parent - The class that called the Address editor.
        attrib - The attribute that is to be edited
        title - The title of the dialog box
        list - list of options for the pop down menu
        """
        EditSecondary.__init__(self, state, uistate, track, attrib, callback)

    def _local_init(self):
        self.top = Glade()
        self.set_window(
            self.top.toplevel, self.top.get_object("title"), _("LDS Ordinance Editor")
        )
        self.share_btn = self.top.get_object("share_place")
        self.add_del_btn = self.top.get_object("add_del_place")

    def _post_init(self):
        self.parents.hide()
        self.parents_label.hide()
        self.parents_select.hide()

    def _connect_signals(self):
        self.define_cancel_button(self.top.get_object("cancel"))
        self.define_help_button(self.top.get_object("help"))
        self.define_ok_button(self.top.get_object("ok"), self.save)

    def _get_types(self):
        return (LdsOrd.SEAL_TO_SPOUSE,)

    def _setup_fields(self):
        self.parents_label = self.top.get_object("parents_label")
        self.parents = self.top.get_object("parents")
        self.parents_select = self.top.get_object("parents_select")

        self.priv = PrivacyButton(
            self.top.get_object("private"), self.obj, self.db.readonly
        )

        self.date_field = MonitoredDate(
            self.top.get_object("date_entry"),
            self.top.get_object("date_stat"),
            self.obj.get_date_object(),
            self.uistate,
            self.track,
            self.db.readonly,
        )

        self.place_field = PlaceEntry(
            self.dbstate,
            self.uistate,
            self.track,
            self.top.get_object("place"),
            self.top.get_object("place_event_box"),
            self.obj.set_place_handle,
            self.obj.get_place_handle,
            self.add_del_btn,
            self.share_btn,
        )

        self.type_menu = MonitoredMenu(
            self.top.get_object("type"),
            self.obj.set_type,
            self.obj.get_type,
            [
                (item[1], item[0])
                for item in LdsOrd._TYPE_MAP
                if item[0] in self._get_types()
            ],
            self.db.readonly,
        )
        self.track_ref_for_deletion("type_menu")

        self.temple_menu = MonitoredStrMenu(
            self.top.get_object("temple"),
            self.obj.set_temple,
            self.obj.get_temple,
            TEMPLES.name_code_data(),
            self.db.readonly,
        )
        self.track_ref_for_deletion("temple_menu")

        self.status_menu = MonitoredMenu(
            self.top.get_object("status"),
            self.obj.set_status,
            self.obj.get_status,
            [
                (item[1], item[0])
                for item in LdsOrd._STATUS_MAP
                if item[0] in _DATA_MAP[self.obj.get_type()]
            ],
            self.db.readonly,
        )
        self.track_ref_for_deletion("status_menu")

    def _create_tabbed_pages(self):
        notebook = Gtk.Notebook()
        self.srcref_list = CitationEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_citation_list(),
            "ldsord_editor_citations",
        )
        self._add_tab(notebook, self.srcref_list)
        self.track_ref_for_deletion("srcref_list")

        self.note_tab = NoteTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_note_list(),
            "ldsord_editor_notes",
            notetype=NoteType.LDS,
        )
        self._add_tab(notebook, self.note_tab)
        self.track_ref_for_deletion("note_tab")

        notebook.show_all()
        self.top.get_object("vbox").pack_start(notebook, True, True, 0)

    def build_menu_names(self, attrib):
        label = _("LDS Ordinance")
        return (label, _("LDS Ordinance Editor"))

    def save(self, *obj):
        """
        Called when the OK button is pressed. Gets data from the
        form and updates the LdsOrd data structure.
        """
        if self.callback:
            self.callback(self.obj)
        self.close()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2008  Brian G. Matherly
# Copyright (C) 2008  Jerome Rapinat
# Copyright (C) 2008  Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# gen.filters.rules/Person/_HasAddress.py

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

from .. import Rule


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Person
from ....db import Database


# -------------------------------------------------------------------------
#
# HasAddress
#
# -------------------------------------------------------------------------
class HasAddress(Rule):
    """Rule that checks for a person with a personal address"""

    labels = [_("Number of instances:"), _("Number must be:")]
    name = _("People with <count> addresses")
    description = _("Matches people with a certain number of personal addresses")
    category = _("General filters")

    def prepare(self, db: Database, user):
        # things we want to do just once, not for every handle
        if self.list[1] == "less than":
            self.count_type = 0
        elif self.list[1] == "greater than":
            self.count_type = 2
        else:
            self.count_type = 1  # "equal to"

        self.userSelectedCount = int(self.list[0])

    def apply_to_one(self, db: Database, person: Person) -> bool:
        count = len(person.address_list)
        if self.count_type == 0:  # "less than"
            return count < self.userSelectedCount
        elif self.count_type == 2:  # "greater than"
            return count > self.userSelectedCount
        # "equal to"
        return count == self.userSelectedCount



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Martin Hawlisch, Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2011       Michiel D. Nauta
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"Import from vCard (RFC 2426)"

# -------------------------------------------------------------------------
#
# standard python modules
#
# -------------------------------------------------------------------------
import sys
import re
import time

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".ImportvCard")

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
ngettext = glocale.translation.ngettext  # else "nearby" comments are ignored
from gramps.gen.errors import GrampsImportError
from gramps.gen.lib import (
    Address,
    Date,
    DateError,
    Event,
    EventRef,
    EventType,
    Name,
    NameType,
    Person,
    Surname,
    Url,
    UrlType,
)
from gramps.gen.db import DbTxn
from gramps.gen.plug.utils import OpenFileOrStdin
from gramps.gen.utils.libformatting import ImportInfo


# -------------------------------------------------------------------------
#
# Support Functions
#
# -------------------------------------------------------------------------
def importData(database, filename, user):
    """Function called by Gramps to import data on persons in vCard format."""
    parser = VCardParser(database)
    try:
        with OpenFileOrStdin(filename) as filehandle:
            parser.parse(filehandle, user)
    except EnvironmentError as msg:
        user.notify_error(_("%s could not be opened\n") % filename, str(msg))
        return
    except GrampsImportError as msg:
        user.notify_error(_("%s could not be opened\n") % filename, str(msg))
        return
    ## a "vCard import report" happens in vCardParser so this is not needed:
    ## (but the imports_test.py unittest currently requires it, so here it is)
    return ImportInfo({_("Results"): _("done")})


def splitof_nameprefix(name):
    """
    Return a (prefix, Surname) tuple by splitting on first uppercase char.

    Shame on Python for not supporting [[:upper:]] in re!
    """
    look_for_capital = False
    for i, char in enumerate(name):
        if look_for_capital:
            if char.isupper():
                return (name[:i].rstrip(), name[i:])
            else:
                look_for_capital = False
        if not char.isalpha():
            look_for_capital = True
    return ("", name)


def fitin(prototype, receiver, element):
    """
    Return the index in string receiver at which element should be inserted
    to match part of prototype.

    Assume that the part of receiver that is not tested does match.
    Don't split to work with lists because element may contain a space!
    Example: fitin("Mr. Gaius Julius Caesar", "Gaius Caesar", "Julius") = 6

    :param prototype: Partly to be matched by inserting element in receiver.
    :type prototype: str
    :param receiver: Space separated words that miss words to match prototype.
    :type receiver: str
    :param element: Words that need to be inserted; error if not in prototype.
    :type element: str
    :returns: Returns index where element fits in receiver, -1 if receiver
              not in prototype, or throws IndexError if element at end receiver.
    :rtype: int
    """
    receiver_idx = 0
    receiver_chunks = receiver.split()
    element_idx = prototype.index(element)
    i = 0
    idx = prototype.find(receiver_chunks[i])
    while idx < element_idx:
        if idx == -1:
            return -1
        receiver_idx += len(receiver_chunks[i]) + 1
        i += 1
        idx = prototype.find(receiver_chunks[i])
    return receiver_idx


# -------------------------------------------------------------------------
#
# VCardParser class
#
# -------------------------------------------------------------------------
class VCardParser:
    """Class to read data in vCard format from a file."""

    DATE_RE = re.compile(r"^(\d{4}-\d{1,2}-\d{1,2})|(?:(\d{4})-?(\d\d)-?(\d\d))")
    GROUP_RE = re.compile(r"^(?:[-0-9A-Za-z]+\.)?(.+)$")  # see RFC 2425 sec5.8.2
    ESCAPE_CHAR = "\\"
    TOBE_ESCAPED = ["\\", ",", ";"]  # order is important
    LINE_CONTINUATION = [" ", "\t"]

    @staticmethod
    def name_value_split(data):
        """Property group.name:value split is on first unquoted colon."""
        colon_idx = data.find(":")
        if colon_idx < 1:
            return ()
        quote_count = data.count('"', 0, colon_idx)
        while quote_count % 2 == 1:
            colon_idx = data.find(":", colon_idx + 1)
            quote_count = data.count('"', 0, colon_idx)
        group_name, value = data[:colon_idx], data[colon_idx + 1 :]
        name_parts = VCardParser.GROUP_RE.match(group_name)
        return (name_parts.group(1), value)

    @staticmethod
    def unesc(data):
        """Remove vCard escape sequences."""
        if type(data) == type("string"):
            for char in reversed(VCardParser.TOBE_ESCAPED):
                data = data.replace(VCardParser.ESCAPE_CHAR + char, char)
            return data
        elif type(data) == type([]):
            return list(map(VCardParser.unesc, data))
        else:
            raise TypeError(
                "vCard unescaping is not implemented for "
                "data type %s." % str(type(data))
            )

    @staticmethod
    def count_escapes(strng):
        """Count the number of escape characters at the end of a string."""
        count = 0
        for char in reversed(strng):
            if char != VCardParser.ESCAPE_CHAR:
                return count
            count += 1
        return count

    @staticmethod
    def split_unescaped(strng, sep):
        """Split on sep if sep is unescaped."""
        strng_parts = strng.split(sep)
        for i in reversed(range(len(strng_parts[:]))):
            if VCardParser.count_escapes(strng_parts[i]) % 2 == 1:
                # the sep was escaped so undo split
                appendix = strng_parts.pop(i + 1)
                strng_parts[i] += sep + appendix
        return strng_parts

    def __init__(self, dbase):
        self.database = dbase
        self.formatted_name = ""
        self.name_parts = ""
        self.next_line = None
        self.trans = None
        self.version = None
        self.person = None
        self.errors = []
        self.number_of_errors = 0

    def __get_next_line(self, filehandle):
        """
        Read and return the line with the next property of the vCard.

        Also if it spans multiple lines (RFC 2425 sec.5.8.1).
        """
        line = self.next_line
        self.next_line = filehandle.readline()
        self.line_num = self.line_num + 1
        while self.next_line and self.next_line[0] in self.LINE_CONTINUATION:
            line = line.rstrip("\n")
            # TODO perhaps next lines superflous because of rU open parameter?
            if len(line) > 0 and line[-1] == "\r":
                line = line[:-1]
            line += self.next_line[1:]
            self.next_line = filehandle.readline()
            self.line_num = self.line_num + 1
        if line:
            line = line.strip()
        else:
            line = None
        return line

    def __add_msg(self, problem, line=None):
        if problem != "":
            self.number_of_errors += 1
        if line:
            message = _("Line %(line)5d: %(prob)s\n") % {"line": line, "prob": problem}
        else:
            message = problem + "\n"
        self.errors.append(message)

    def parse(self, filehandle, user):
        """
        Prepare the database and parse the input file.

        :param filehandle: open file handle positioned at start of the file
        """
        tym = time.time()
        self.person = None
        self.database.disable_signals()
        with DbTxn(_("vCard import"), self.database, batch=True) as self.trans:
            self._parse_vCard_file(filehandle)
        self.database.enable_signals()
        self.database.request_rebuild()
        tym = time.time() - tym
        # Translators: leave all/any {...} untranslated
        msg = ngettext(
            "Import Complete: {number_of} second",
            "Import Complete: {number_of} seconds",
            tym,
        ).format(number_of=tym)
        LOG.debug(msg)
        if self.number_of_errors == 0:
            message = _("vCard import report: No errors detected")
        else:
            message = (
                _("vCard import report: %s errors detected\n") % self.number_of_errors
            )
        if hasattr(user.uistate, "window"):
            parent_window = user.uistate.window
        else:
            parent_window = None
        user.info(message, "".join(self.errors), parent=parent_window, monospaced=True)

    def _parse_vCard_file(self, filehandle):
        """Read each line of the input file and act accordingly."""
        self.next_line = filehandle.readline()
        self.line_num = 1

        while True:
            line = self.__get_next_line(filehandle)
            if line is None:
                break
            if line == "":
                continue

            if line.find(":") == -1:
                continue
            line_parts = self.name_value_split(line)
            if not line_parts:
                continue

            # No check for escaped ; because only fields[0] is used.
            fields = line_parts[0].split(";")

            property_name = fields[0].upper()
            if property_name == "BEGIN":
                self.next_person()
            elif property_name == "END":
                self.finish_person()
            elif property_name == "VERSION":
                self.check_version(fields, line_parts[1])
            elif property_name == "FN":
                self.add_formatted_name(fields, line_parts[1])
            elif property_name == "N":
                self.add_name_parts(fields, line_parts[1])
            elif property_name == "NICKNAME":
                self.add_nicknames(fields, line_parts[1])
            elif property_name == "SORT-STRING":
                self.add_sortas(fields, line_parts[1])
            elif property_name == "ADR":
                self.add_address(fields, line_parts[1])
            elif property_name == "TEL":
                self.add_phone(fields, line_parts[1])
            elif property_name == "BDAY":
                self.add_birthday(fields, line_parts[1])
            elif property_name == "ROLE":
                self.add_occupation(fields, line_parts[1])
            elif property_name == "URL":
                self.add_url(fields, line_parts[1])
            elif property_name == "EMAIL":
                self.add_email(fields, line_parts[1])
            elif property_name == "X-GENDER" or property_name == "GENDER":
                # vCard 3.0 only has X-GENDER, GENDER is 4.0 syntax,
                # but we want to be robust here.
                self.add_gender(fields, line_parts[1])
            elif property_name == "PRODID":
                # Included cause vCards made by Gramps have this prop.
                pass
            else:
                self.__add_msg(
                    _("Token >%(token)s< unknown. line skipped: %(line)s")
                    % {"token": (fields[0], line), "line": self.line_num - 1}
                )

    def finish_person(self):
        """All info has been collected, write to database."""
        if self.person is not None:
            if self.add_name():
                self.database.add_person(self.person, self.trans)
        self.person = None

    def next_person(self):
        """A vCard for another person is started."""
        if self.person is not None:
            self.finish_person()
            self.__add_msg(
                _(
                    "BEGIN property not properly closed by END "
                    "property, Gramps can't cope with nested vCards."
                ),
                self.line_num - 1,
            )
        self.person = Person()
        self.formatted_name = ""
        self.name_parts = ""

    def check_version(self, fields, data):
        """Check the version of the vCard, only version 3.0 is supported."""
        self.version = data
        if self.version != "3.0":
            raise GrampsImportError(
                _("Import of vCards version %s is " "not supported by Gramps.")
                % self.version
            )

    def add_formatted_name(self, fields, data):
        """Read the FN property of a vCard."""
        if not self.formatted_name:
            self.formatted_name = self.unesc(str(data)).strip()

    def add_name_parts(self, fields, data):
        """Read the N property of a vCard."""
        if not self.name_parts:
            self.name_parts = data.strip()

    def add_name(self):
        """
        Add the name to the person.

        Returns True on success, False on failure.
        """
        if not self.name_parts.strip():
            self.__add_msg(
                _(
                    "The vCard is malformed. It is missing the compulsory N "
                    "property, so there is no name; skip it."
                ),
                self.line_num - 1,
            )
            return False
        if not self.formatted_name:
            self.__add_msg(
                _(
                    "The vCard is malformed. It is missing the compulsory FN "
                    "property, get name from N alone."
                ),
                self.line_num - 1,
            )
        data_fields = self.split_unescaped(self.name_parts, ";")
        if len(data_fields) != 5:
            self.__add_msg(
                _("The vCard is malformed. Wrong number of name " "components."),
                self.line_num - 1,
            )

        name = Name()
        name.set_type(NameType(NameType.BIRTH))

        if data_fields[0].strip():
            # assume first surname is primary
            for surname_str in self.split_unescaped(data_fields[0], ","):
                surname = Surname()
                prefix, sname = splitof_nameprefix(self.unesc(surname_str))
                surname.set_surname(sname.strip())
                surname.set_prefix(prefix.strip())
                name.add_surname(surname)
            name.set_primary_surname()

        if len(data_fields) > 1 and data_fields[1].strip():
            given_name = " ".join(self.unesc(self.split_unescaped(data_fields[1], ",")))
        else:
            given_name = ""
        if len(data_fields) > 2 and data_fields[2].strip():
            additional_names = " ".join(
                self.unesc(self.split_unescaped(data_fields[2], ","))
            )
        else:
            additional_names = ""
        self.add_firstname(given_name.strip(), additional_names.strip(), name)

        if len(data_fields) > 3 and data_fields[3].strip():
            name.set_title(
                " ".join(self.unesc(self.split_unescaped(data_fields[3], ",")))
            )
        if len(data_fields) > 4 and data_fields[4].strip():
            name.set_suffix(
                " ".join(self.unesc(self.split_unescaped(data_fields[4], ",")))
            )

        self.person.set_primary_name(name)
        return True

    def add_firstname(self, given_name, additional_names, name):
        """
        Combine given_name and additional_names and add as firstname to name.

        If possible try to add given_name as call name.
        """
        default = "%s %s" % (given_name, additional_names)
        if self.formatted_name:
            if given_name:
                if additional_names:
                    given_name_pos = self.formatted_name.find(given_name)
                    if given_name_pos != -1:
                        add_names_pos = self.formatted_name.find(additional_names)
                        if add_names_pos != -1:
                            if given_name_pos <= add_names_pos:
                                firstname = default
                                # Uncertain if given name is used as callname
                            else:
                                firstname = "%s %s" % (additional_names, given_name)
                                name.set_call_name(given_name)
                        else:
                            idx = fitin(
                                self.formatted_name, additional_names, given_name
                            )
                            if idx == -1:
                                # Additional names is not in formatted name
                                firstname = default
                            else:  # Given name in middle of additional names
                                firstname = "%s%s %s" % (
                                    additional_names[:idx],
                                    given_name,
                                    additional_names[idx:],
                                )
                                name.set_call_name(given_name)
                    else:  # Given name is not in formatted name
                        firstname = default
                else:  # There are no additional_names
                    firstname = given_name
            else:  # There is no given_name
                firstname = additional_names
        else:  # There is no formatted name
            firstname = default
        name.set_first_name(firstname.strip())
        return

    def add_nicknames(self, fields, data):
        """Read the NICKNAME property of a vCard."""
        for nick in self.split_unescaped(data, ","):
            nickname = nick.strip()
            if nickname:
                name = Name()
                name.set_nick_name(self.unesc(nickname))
                self.person.add_alternate_name(name)

    def add_sortas(self, fields, data):
        """Read the SORT-STRING property of a vCard."""
        # TODO
        pass

    def add_address(self, fields, data):
        """Read the ADR property of a vCard."""
        data_fields = self.split_unescaped(data, ";")
        data_fields = [x.strip() for x in self.unesc(data_fields)]
        if "".join(data_fields):
            addr = Address()

            def add_street(strng):
                if strng:
                    already = addr.get_street()
                    if already:
                        addr.set_street("%s %s" % (already, strng))
                    else:
                        addr.set_street(strng)

            addr._add_street = add_street
            set_func = [
                "_add_street",
                "_add_street",
                "_add_street",
                "set_city",
                "set_state",
                "set_postal_code",
                "set_country",
            ]
            for i, data in enumerate(data_fields):
                if i >= len(set_func):
                    break
                getattr(addr, set_func[i])(data)
            self.person.add_address(addr)

    def add_phone(self, fields, data):
        """Read the TEL property of a vCard."""
        tel = data.strip()
        if tel:
            addr = Address()
            addr.set_phone(self.unesc(tel))
            self.person.add_address(addr)

    def add_birthday(self, fields, data):
        """Read the BDAY property of a vCard."""
        date_str = data.strip()
        date_match = VCardParser.DATE_RE.match(date_str)
        date = Date()
        if date_match:
            if date_match.group(2):
                date_str = "%s-%s-%s" % (
                    date_match.group(2),
                    date_match.group(3),
                    date_match.group(4),
                )
            else:
                date_str = date_match.group(1)
            y, m, d = [int(x, 10) for x in date_str.split("-")]
            try:
                date.set(value=(d, m, y, False))
            except DateError:
                # Translators: leave the {vcard_snippet} untranslated
                # in the format string, but you may re-order it if needed.
                self.__add_msg(
                    _(
                        "Invalid date in BDAY {vcard_snippet}, "
                        "preserving date as text."
                    ).format(vcard_snippet=data),
                    self.line_num - 1,
                )
                date.set(modifier=Date.MOD_TEXTONLY, text=data)
        else:
            if date_str:
                # Translators: leave the {vcard_snippet} untranslated.
                self.__add_msg(
                    _(
                        "Date {vcard_snippet} not in appropriate format "
                        "yyyy-mm-dd, preserving date as text."
                    ).format(vcard_snippet=date_str),
                    self.line_num - 1,
                )
                date.set(modifier=Date.MOD_TEXTONLY, text=date_str)
            else:  # silently ignore an empty BDAY record
                return
        event = Event()
        event.set_type(EventType(EventType.BIRTH))
        event.set_date_object(date)
        self.database.add_event(event, self.trans)

        event_ref = EventRef()
        event_ref.set_reference_handle(event.get_handle())
        self.person.set_birth_ref(event_ref)

    def add_occupation(self, fields, data):
        """Read the ROLE property of a vCard."""
        occupation = data.strip()
        if occupation:
            event = Event()
            event.set_type(EventType(EventType.OCCUPATION))
            event.set_description(self.unesc(occupation))
            self.database.add_event(event, self.trans)

            event_ref = EventRef()
            event_ref.set_reference_handle(event.get_handle())
            self.person.add_event_ref(event_ref)

    def add_url(self, fields, data):
        """Read the URL property of a vCard."""
        href = data.strip()
        if href:
            url = Url()
            url.set_path(self.unesc(href))
            self.person.add_url(url)

    def add_email(self, fields, data):
        """Read the EMAIL property of a vCard."""
        email = data.strip()
        if email:
            url = Url()
            url.set_type(UrlType(UrlType.EMAIL))
            url.set_path(self.unesc(email))
            self.person.add_url(url)

    def add_gender(self, fields, data):
        """Read the GENDER property of a vCard."""
        gender_value = data.strip()
        if gender_value:
            gender_value = gender_value.upper()
            gender_value = gender_value[0]
            if gender_value == "M":
                gender = Person.MALE
            elif gender_value == "F":
                gender = Person.FEMALE
            elif gender_value == "O":
                gender = Person.OTHER
            else:
                return
            self.person.set_gender(gender)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import unittest

from ..callback import Callback


class TestCallback(unittest.TestCase):
    def test_simple(self):
        class TestSignals(Callback):
            __signals__ = {"test-signal": (int,)}

        rl = []

        def fn(i, r=rl):
            rl.append(i)

        t = TestSignals()
        t.connect("test-signal", fn)
        t.emit("test-signal", (1,))

        self.assertEqual(len(rl), 1, "No signal emitted")
        self.assertEqual(rl[0], 1, "Wrong argument recieved")

    def test_exception_catch(self):
        class TestSignals(Callback):
            __signals__ = {"test-signal": (int,)}

        rl = []

        def fn(i, r=rl):
            rl.append(i)

        def borked(i):
            """
            This intentionally raises a NameError exception
            """
            with self.assertRaises(NameError):
                rubish.append(i)

        t = TestSignals()

        t.connect("test-signal", borked)
        t.connect("test-signal", fn)
        t.emit("test-signal", (1,))

        self.assertEqual(len(rl), 1, "No signal emitted")
        self.assertEqual(rl[0], 1, "Wrong argument received")

    def test_disconnect(self):
        class TestSignals(Callback):
            __signals__ = {"test-signal": (int,)}

        rl = []

        def fn(i, r=rl):
            rl.append(i)

        t = TestSignals()
        key = t.connect("test-signal", fn)
        t.emit("test-signal", (1,))

        self.assertEqual(len(rl), 1, "No signal emitted")
        self.assertEqual(rl[0], 1, "Wrong argument recieved")

        t.disconnect(key)

        t.emit("test-signal", (1,))

        self.assertEqual(len(rl), 1, "Callback not disconnected")
        self.assertEqual(rl[0], 1, "Callback not disconnected")

    def test_noargs(self):
        class TestSignals(Callback):
            __signals__ = {"test-noargs": None}

        rl = []

        def fn(r=rl):
            rl.append(1)

        t = TestSignals()
        t.connect("test-noargs", fn)
        t.emit("test-noargs")

        self.assertEqual(len(rl), 1, "No signal emitted")
        self.assertEqual(rl[0], 1, "Wrong argument recieved")

    def test_no_callback(self):
        class TestSignals(Callback):
            __signals__ = {"test-noargs": None}

        t = TestSignals()
        t.emit("test-noargs")

    def test_subclassing(self):
        class TestSignals(Callback):
            __signals__ = {"test-signal": (int,)}

        class TestSignalsSubclass(TestSignals):
            __signals__ = {
                "test-sub-signal": (int,),
            }

        rl = []

        def fn(i, r=rl):
            rl.append(i)

        t = TestSignalsSubclass()
        t.connect("test-signal", fn)
        t.emit("test-signal", (1,))

        self.assertEqual(len(rl), 1, "No signal emitted")
        self.assertEqual(rl[0], 1, "Wrong argument recieved")

        t.connect("test-sub-signal", fn)
        t.emit("test-sub-signal", (1,))

        self.assertEqual(len(rl), 2, "No subclass signal emitted")
        self.assertEqual(rl[1], 1, "Wrong argument recieved in subclass")

    def test_signal_block(self):
        class TestSignals(Callback):
            __signals__ = {"test-signal": (int,)}

        rl = []

        def fn(i, r=rl):
            rl.append(i)

        t = TestSignals()
        t.connect("test-signal", fn)
        t.emit("test-signal", (1,))

        self.assertEqual(len(rl), 1, "No signal emitted")
        self.assertEqual(rl[0], 1, "Wrong argument recieved")

        Callback.disable_all_signals()
        t.emit("test-signal", (1,))
        self.assertEqual(len(rl), 1, "Signal emitted while class blocked")

        Callback.enable_all_signals()
        t.emit("test-signal", (1,))
        self.assertEqual(len(rl), 2, "Signals not class unblocked")

        t.disable_signals()
        t.emit("test-signal", (1,))
        self.assertEqual(len(rl), 2, "Signal emitted while instance blocked")

        t.enable_signals()
        t.emit("test-signal", (1,))
        self.assertEqual(len(rl), 3, "Signals not instance unblocked")

    def test_type_checking(self):
        class TestSignals(Callback):
            __signals__ = {
                "test-int": (int,),
                "test-list": (list,),
                "test-object": (object,),
                "test-str": (str,),
                "test-float": (float,),
                "test-dict": (dict,),
                "test-lots": (int, str, list, object, float),
            }

        rl = []

        def fn(i, r=rl):
            rl.append(i)

        t = TestSignals()
        t.connect("test-int", fn), t.emit("test-int", (1,))
        self.assertIsInstance(rl[0], int, "not int")

        t.connect("test-list", fn), t.emit("test-list", ([1, 2],))
        self.assertIsInstance(rl[1], list, "not list")

        t.connect("test-object", fn), t.emit("test-object", (t,))
        self.assertIsInstance(rl[2], object, "not object")

        t.connect("test-float", fn), t.emit("test-float", (2.3,))
        self.assertIsInstance(rl[3], float, "not float")

        t.connect("test-dict", fn), t.emit("test-dict", ({1: 2},))
        self.assertIsInstance(rl[4], dict, "not dict")

        rl = []

        def fn2(i, s, l, o, f, r=rl):
            rl.append(i)

        t.connect("test-lots", fn2), t.emit("test-lots", (1, "a", [1, 2], t, 1.2))
        self.assertIsInstance(rl[0], int, "not lots")

        # This should fail because the type of arg1 is wrong
        res = []

        def fn3(s, r=res):
            res.append(s)

        t._warn = fn3
        t.connect("test-lots", fn2), t.emit("test-lots", ("a", "a", [1, 2], t, 1.2))
        self.assertEqual(res[0][0:6], "Signal", "Type error not detected")

    def test_recursion_block(self):
        class TestSignals(Callback):
            __signals__ = {"test-recursion": (Callback,)}

        def fn(cb):
            cb.emit("test-recursion", (t,))

        res = []

        def fn2(s, r=res):
            res.append(s)

        t = TestSignals()
        t._warn = fn2
        t.connect("test-recursion", fn)

        try:
            t.emit("test-recursion", (t,))
        except RuntimeError:
            self.fail("signal recursion not blocked1.")

        self.assertEqual(res[0][0:6], "Signal", "signal recursion not blocked")

    def test_multisignal_recursion_block(self):
        class TestSignals(Callback):
            __signals__ = {
                "test-top": (Callback,),
                "test-middle": (Callback,),
                "test-bottom": (Callback,),
            }

        def top(cb):
            cb.emit("test-middle", (t,))

        def middle(cb):
            cb.emit("test-bottom", (t,))

        def bottom(cb):
            cb.emit("test-top", (t,))

        res = []

        def fn2(s, r=res):
            res.append(s)

        t = TestSignals()
        t._warn = fn2
        t.connect("test-top", top)
        t.connect("test-middle", middle)
        t.connect("test-bottom", bottom)

        try:
            t.emit("test-top", (t,))
        except RuntimeError:
            self.fail("multisignal recursion not blocked1.")

        self.assertEqual(res[0][0:6], "Signal", "multisignal recursion not blocked")


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2013  Vassilii Khachaturov
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import unittest

from .. import _datestrings
from ...lib.date import Date


class DateStringsTest(unittest.TestCase):
    def setUp(self):
        from ...utils.grampslocale import GrampsLocale

        self.ds = _datestrings.DateStrings(GrampsLocale())  # whatever the default...
        self.ds_EN = _datestrings.DateStrings(GrampsLocale(languages="en"))
        self.ds_RU = _datestrings.DateStrings(GrampsLocale(languages="ru"))

    def testTwelfthMonthIsDecember(self):
        self.assertEqual(self.ds_EN.long_months[12], "December")
        self.assertEqual(self.ds_EN.short_months[12], "Dec")

    # May is 3-letter in Russian, and so abbreviated form
    # will be different for inflections!
    def testRussianHasDifferentInflectionsForShortMay(self):
        v5 = list(self.ds_RU.short_months[5].variants())
        self.assertTrue(len(v5) > 1, msg=v5)

    def testEnAdarI_in_AdarII(self):
        adar1 = self.ds_EN.hebrew[6]
        adar2 = self.ds_EN.hebrew[7]
        self.assertIn(str(adar1), str(adar2))

    def testEnLastFrenchIsExtra(self):
        self.assertEqual(str(self.ds_EN.french[-1]), "Extra")

    def testEnPersianKhordadMordad(self):
        khordad = self.ds_EN.persian[3].lower()
        mordad = self.ds_EN.persian[5].lower()
        self.assertEqual(khordad, "khordad")
        self.assertEqual(mordad, "mordad")

    def testEnIslamicRamadan9(self):
        self.assertEqual(str(self.ds_EN.islamic[9]), "Ramadan")

    def testFirstStringEmpty(self):
        self.assertEqual(self.ds.long_months[0], "")
        self.assertEqual(self.ds.short_months[0], "")
        self.assertEqual(self.ds.alt_long_months[0], "")
        self.assertEqual(self.ds.long_days[0], "")

    def testCalendarIndex(self):
        self.assertEqual(self.ds_EN.calendar[Date.CAL_GREGORIAN], "Gregorian")
        self.assertEqual(self.ds_EN.calendar[Date.CAL_JULIAN], "Julian")
        self.assertEqual(self.ds_EN.calendar[Date.CAL_HEBREW], "Hebrew")
        self.assertEqual(self.ds_EN.calendar[Date.CAL_FRENCH], "French Republican")
        self.assertEqual(self.ds_EN.calendar[Date.CAL_PERSIAN], "Persian")
        self.assertEqual(self.ds_EN.calendar[Date.CAL_ISLAMIC], "Islamic")
        self.assertEqual(self.ds_EN.calendar[Date.CAL_SWEDISH], "Swedish")

    def testDayNamesLenIs8(self):
        self.assertEqual(len(self.ds.long_days), 8)


if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007  B. Malengier
# Copyright (C) 2008  Brian G. Matherly
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#

"""
Display a person's relations to the home person
"""
# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

from gramps.gen.simple import SimpleAccess, SimpleDoc
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.relationship import get_relationship_calculator

# define the formatting string once as a constant. Since this is reused

_FMT = "%-3d %s"
_FMT_VOID = "    %s"
_FMT_DET1 = "%-3s %-15s"
_FMT_DET2 = "%-30s %-15s\t%-10s %-2s"


def run(database, document, person):
    """
    Create the report class, and produce the quick report
    """
    report = AllRelReport(database, document, person)
    report.run()


class AllRelReport:
    """
    Obtains all relationships, displays the relations, and in details, the
    relation path
    """

    def __init__(self, database, document, person):
        self.database = database
        self.person = person
        self.sdb = SimpleAccess(database)
        self.sdoc = SimpleDoc(document)
        self.rel_class = get_relationship_calculator(glocale)

        self.msg_list = []

    def run(self):
        # get home_person
        self.home_person = self.database.get_default_person()
        if not self.home_person:
            self.sdoc.paragraph(_("Home person not set."))
            return

        self.print_title()

        p2 = self.sdb.name(self.home_person)
        p1 = self.sdb.name(self.person)
        if self.person.handle == self.home_person.handle:
            self.sdoc.paragraph(
                _FMT_VOID
                % (_("%(person)s and %(active_person)s are the same person."))
                % {"person": p1, "active_person": p2}
            )
            return

        # check if not a family too:
        is_spouse = self.rel_class.is_spouse(
            self.database, self.home_person, self.person
        )
        if is_spouse:
            rel_string = is_spouse
            rstr = _("%(person)s is the %(relationship)s of %(active_person)s.") % {
                "person": p1,
                "relationship": rel_string,
                "active_person": p2,
            }
            self.sdoc.paragraph(_FMT_VOID % (rstr))
            self.sdoc.paragraph("")

        # obtain all relationships, assume home person has largest tree
        common, self.msg_list = self.rel_class.get_relationship_distance_new(
            self.database,
            self.person,
            self.home_person,
            all_families=True,
            all_dist=True,
            only_birth=False,
        )
        # all relations
        if (not common or common[0][0] == -1) and not is_spouse:
            rstr = _(
                "%(person)s and %(active_person)s are not " "directly related."
            ) % {"person": p2, "active_person": p1}
            self.sdoc.paragraph(_FMT_VOID % (rstr))
            self.sdoc.paragraph("")

        # collapse common so parents of same fam in common are one line
        commonnew = self.rel_class.collapse_relations(common)
        self.print_details_header(
            commonnew, self.home_person, self.person, skip_list_text=None
        )
        self.print_details_path(commonnew, self.home_person, self.person)
        self.print_details_path(commonnew, self.home_person, self.person, first=False)

        if not common or common[0][0] == -1:
            self.remarks(self.msg_list)
            self.msg_list = []
            # check inlaw relation next
        else:
            # stop
            return

        # we check the inlaw relationships if not partners.
        if is_spouse:
            return
        handles_done = [(self.person.handle, self.home_person.handle)]
        inlaws_pers = [self.person] + self.get_inlaws(self.person)
        inlaws_home = [self.home_person] + self.get_inlaws(self.home_person)
        # remove overlap:
        inlaws_home = [x for x in inlaws_home if x not in inlaws_pers]
        inlawwritten = False
        skiplist = []
        commonnew = []
        for inlawpers in inlaws_pers:
            for inlawhome in inlaws_home:
                if (inlawpers, inlawhome) in handles_done:
                    continue
                else:
                    handles_done.append((inlawpers, inlawhome))
                common, msg = self.rel_class.get_relationship_distance_new(
                    self.database,
                    inlawpers,
                    inlawhome,
                    all_families=True,
                    all_dist=True,
                    only_birth=False,
                )
                if msg:
                    self.msg_list += msg
                if common and not common[0][0] == -1:
                    if not inlawwritten:
                        rstr = _(
                            "%(person)s and %(active_person)s have "
                            "following in-law relations:"
                        ) % {"person": p2, "active_person": p1}
                        self.sdoc.paragraph(_FMT_VOID % (rstr))
                        self.sdoc.paragraph("")
                        inlawwritten = True
                else:
                    continue
                inlawb = not inlawpers.handle == self.person.handle
                inlawa = not inlawhome.handle == self.home_person.handle
                commonnew.append(
                    (
                        inlawa,
                        inlawb,
                        inlawhome,
                        inlawpers,
                        self.rel_class.collapse_relations(common),
                    )
                )
        skip = []
        skip_text = []
        count = 1
        for inlawa, inlawb, inlawhome, inlawpers, commonrel in commonnew:
            count = self.print_details_header(
                commonrel,
                inlawhome,
                inlawpers,
                inlawa=inlawa,
                inlawb=inlawb,
                count=count,
                skip_list=skip,
                skip_list_text=skip_text,
            )
        count = 1
        for inlawa, inlawb, inlawhome, inlawpers, commonrel in commonnew:
            self.print_details_path(
                commonrel,
                inlawhome,
                inlawpers,
                inlawa=inlawa,
                inlawb=inlawb,
                count=count,
                skip_list=skip,
            )
            count = self.print_details_path(
                commonrel,
                inlawhome,
                inlawpers,
                inlawa=inlawa,
                inlawb=inlawb,
                count=count,
                skip_list=skip,
                first=False,
            )
        self.remarks(self.msg_list, True)

    def get_inlaws(self, person):
        inlaws = []
        family_handles = person.get_family_handle_list()
        for handle in family_handles:
            fam = self.database.get_family_from_handle(handle)
            if fam.father_handle and not fam.father_handle == person.handle:
                inlaws.append(self.database.get_person_from_handle(fam.father_handle))
            elif fam.mother_handle and not fam.mother_handle == person.handle:
                inlaws.append(self.database.get_person_from_handle(fam.mother_handle))
        return inlaws

    def print_title(self):
        """print the title"""
        p2 = self.sdb.name(self.home_person)
        p1 = self.sdb.name(self.person)
        self.sdoc.title(
            _("Relationships of %(person)s to %(active_person)s")
            % {"person": p1, "active_person": p2}
        )
        self.sdoc.paragraph("")

    def print_details_header(
        self,
        relations,
        pers1,
        pers2,
        inlawa=False,
        inlawb=False,
        count=1,
        skip_list=[],
        skip_list_text=[],
    ):
        if not relations or relations[0][0] == -1:
            return count

        sdoc = self.sdoc
        rel_class = self.rel_class
        for relation in relations:
            birth = self.rel_class.only_birth(
                relation[2]
            ) and self.rel_class.only_birth(relation[4])
            distorig = len(relation[4])
            distother = len(relation[2])
            if distorig == distother == 1 and not inlawa and not inlawb:
                rel_str = self.rel_class.get_sibling_relationship_string(
                    self.rel_class.get_sibling_type(self.database, pers1, pers2),
                    self.home_person.get_gender(),
                    self.person.get_gender(),
                )
            else:
                rel_str = self.rel_class.get_single_relationship_string(
                    distorig,
                    distother,
                    self.home_person.get_gender(),
                    self.person.get_gender(),
                    relation[4],
                    relation[2],
                    only_birth=birth,
                    in_law_a=inlawa,
                    in_law_b=inlawb,
                )
            if skip_list_text is not None:
                if rel_str in skip_list_text:
                    skip_list.append(count)
                else:
                    skip_list_text.append(rel_str)
                    sdoc.paragraph(_FMT % (count - len(skip_list), rel_str))
            else:
                sdoc.paragraph(_FMT % (count, rel_str))
            count += 1
        return count

    def print_details_path(
        self,
        relations,
        pers1,
        pers2,
        inlawa=False,
        inlawb=False,
        count=1,
        skip_list=[],
        first=True,
    ):
        if not relations or relations[0][0] == -1:
            return count

        sdoc = self.sdoc
        rel_class = self.rel_class
        p2 = self.sdb.name(self.home_person)
        p1 = self.sdb.name(self.person)
        pers = p2
        inlaw = inlawa
        if first:
            pers = p1
            inlaw = inlawb

        if count == 1:
            sdoc.paragraph("")
            sdoc.header1(
                _("Detailed path from %(person)s to common ancestor") % {"person": pers}
            )
            sdoc.paragraph("")
            sdoc.header2(_FMT_DET1 % ("   ", _("Name Common ancestor")))
            sdoc.header2(_FMT_DET2 % (" ", _("Parent"), _("Birth"), _("Family")))
            sdoc.paragraph("")
        for relation in relations:
            if count in skip_list:
                count += 1
                continue
            counter = str(count - len([x for x in range(count) if x + 1 in skip_list]))
            name = _("Unknown")
            if relation[1]:
                name = self.sdb.name(
                    self.database.get_person_from_handle(relation[1][0])
                )
                for handle in relation[1][1:]:
                    name += (
                        " "
                        + _("and")
                        + " "
                        + self.sdb.name(self.database.get_person_from_handle(handle))
                    )
            sdoc.paragraph(_FMT_DET1 % (counter, name))
            if inlaw:
                sdoc.paragraph(_FMT_DET2 % (" ", _("Partner"), " ", " "))
            if first:
                ind1 = 2
                ind2 = 3
            else:
                ind1 = 4
                ind2 = 5
            for rel, fam in zip(relation[ind1], relation[ind2]):
                par_str = _("Unknown")  # when sibling, parent is unknown
                if rel == rel_class.REL_MOTHER or rel == rel_class.REL_MOTHER_NOTBIRTH:
                    par_str = _("Mother")
                if rel == rel_class.REL_FATHER or rel == rel_class.REL_FATHER_NOTBIRTH:
                    par_str = _("Father")
                if (
                    rel == rel_class.REL_FAM_BIRTH
                    or rel == rel_class.REL_FAM_NONBIRTH
                    or rel == rel_class.REL_FAM_BIRTH_MOTH_ONLY
                    or rel == rel_class.REL_FAM_BIRTH_FATH_ONLY
                ):
                    par_str = _("Parents")
                birth_str = _("Yes")
                if (
                    rel == rel_class.REL_MOTHER_NOTBIRTH
                    or rel == rel_class.REL_FATHER_NOTBIRTH
                    or rel == rel_class.REL_FAM_NONBIRTH
                ):
                    birth_str = _("No")
                elif (
                    rel == rel_class.REL_FAM_BIRTH_FATH_ONLY
                    or rel == rel_class.REL_FAM_BIRTH_MOTH_ONLY
                ):
                    birth_str = _("Partial")
                famstr = ""
                if isinstance(fam, list):
                    famstr = str(fam[0] + 1)
                    for val in fam[1:]:
                        # TODO for Arabic, should the next comma be translated?
                        famstr += ", " + str(val + 1)
                elif fam is not None:
                    famstr = str(fam + 1)
                sdoc.paragraph(_FMT_DET2 % (" ", par_str, birth_str, famstr))
                counter = ""
                name = ""
            count += 1
        return count

    def remarks(self, msg_list, inlaw=False):
        if msg_list:
            sdoc = self.sdoc
            sdoc.paragraph("")
            if inlaw:
                sdoc.header1(_("Remarks with inlaw family"))
            else:
                sdoc.header1(_("Remarks"))
            sdoc.paragraph("")
            sdoc.paragraph(_("The following problems were encountered:"))

            list(map(sdoc.paragraph, msg_list))
            sdoc.paragraph("")
            sdoc.paragraph("")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2009       Douglas S. Blank
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Date editing module for Gramps.

The EditDate provides visual feedback to the user to indicate if
the associated GtkEntry box contains a valid date.
Red means that the date is not valid, and will be viewed as a text string
instead of a date.

The DateEditor provides a dialog in which the date can be
unambiguously built using UI controls such as menus and spin buttons.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------


# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
import logging

# -------------------------------------------------------------------------
#
# GNOME modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.config import config
from gramps.gen.lib.date import Date, DateError, calendar_has_fixed_newyear
from gramps.gen.datehandler import displayer
from gramps.gen.const import URL_MANUAL_SECT1
from ..display import display_help
from ..managedwindow import ManagedWindow
from ..glade import Glade

LOG = logging.getLogger(".EditDate")
_ = glocale.translation.sgettext
# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
MOD_TEXT = (
    (Date.MOD_NONE, _("Regular")),
    (Date.MOD_BEFORE, _("Before")),
    (Date.MOD_AFTER, _("After")),
    (Date.MOD_ABOUT, _("About")),
    (Date.MOD_RANGE, _("Range")),
    (Date.MOD_FROM, _("From")),
    (Date.MOD_TO, _("To")),
    (Date.MOD_SPAN, _("Span")),
    (Date.MOD_TEXTONLY, _("Text only")),
)

QUAL_TEXT = (
    (Date.QUAL_NONE, _("Regular")),
    (Date.QUAL_ESTIMATED, _("Estimated")),
    (Date.QUAL_CALCULATED, _("Calculated")),
)

CAL_TO_MONTHS_NAMES = {
    Date.CAL_GREGORIAN: displayer.short_months,
    Date.CAL_JULIAN: displayer.short_months,
    Date.CAL_HEBREW: displayer.hebrew,
    Date.CAL_FRENCH: displayer.french,
    Date.CAL_PERSIAN: displayer.persian,
    Date.CAL_ISLAMIC: displayer.islamic,
    Date.CAL_SWEDISH: displayer.swedish,
}

WIKI_HELP_PAGE = URL_MANUAL_SECT1
WIKI_HELP_SEC = _("Editing_dates", "manual")


# -------------------------------------------------------------------------
#
# EditDate
#
# -------------------------------------------------------------------------
class EditDate(ManagedWindow):
    """
    Dialog allowing to build the date precisely, to correct possible
    limitations of parsing and/or underlying structure of :class:`.Date`.
    """

    def __init__(self, date, uistate, track):
        """
        Initiate and display the dialog.
        """
        ManagedWindow.__init__(self, uistate, track, self)

        # Create self.date as a copy of the given Date object.
        self.date = Date(date)

        self.top = Glade()

        self.set_window(
            self.top.toplevel, self.top.get_object("title"), _("Date selection")
        )
        self.setup_configs("interface.editdate", 620, 320)

        self.statusbar = self.top.get_object("statusbar")
        self.ok_button = self.top.get_object("ok_button")
        self.calendar_box = self.top.get_object("calendar_box")
        for name in Date.ui_calendar_names:
            self.calendar_box.get_model().append([name])

        self.new_year = self.top.get_object("newyear")
        self.new_year.set_text(self.date.newyear_to_str())

        cal = self.date.get_calendar()
        self.calendar_box.set_active(cal)
        self.align_newyear_ui_with_calendar(cal)
        self.calendar_box.connect("changed", self.switch_calendar)

        self.quality_box = self.top.get_object("quality_box")
        for item_number, item in enumerate(QUAL_TEXT):
            self.quality_box.append_text(item[1])
            if self.date.get_quality() == item[0]:
                self.quality_box.set_active(item_number)

        self.type_box = self.top.get_object("type_box")
        for item_number, item in enumerate(MOD_TEXT):
            self.type_box.append_text(item[1])
            if self.date.get_modifier() == item[0]:
                self.type_box.set_active(item_number)
        self.type_box.connect("changed", self.switch_type)

        self.start_month_box = self.top.get_object("start_month_box")
        self.stop_month_box = self.top.get_object("stop_month_box")
        month_names = CAL_TO_MONTHS_NAMES[self.date.get_calendar()]
        for name in month_names:
            self.start_month_box.append_text(name)
            self.stop_month_box.append_text(name)
        self.start_month_box.set_active(self.date.get_month())
        self.stop_month_box.set_active(self.date.get_stop_month())

        self.start_day = self.top.get_object("start_day")
        self.start_day.set_value(self.date.get_day())
        self.start_year = self.top.get_object("start_year")
        self.start_year.set_value(self.date.get_year())

        self.stop_day = self.top.get_object("stop_day")
        self.stop_day.set_value(self.date.get_stop_day())
        self.stop_year = self.top.get_object("stop_year")
        self.stop_year.set_value(self.date.get_stop_year())

        self.dual_dated = self.top.get_object("dualdated")

        # Disable second date controls if not compound date
        if not self.date.is_compound():
            self.stop_day.set_sensitive(0)
            self.stop_month_box.set_sensitive(0)
            self.stop_year.set_sensitive(0)

        # Disable the rest of controls if a text-only date
        if self.date.get_modifier() == Date.MOD_TEXTONLY:
            self.start_day.set_sensitive(0)
            self.start_month_box.set_sensitive(0)
            self.start_year.set_sensitive(0)
            self.calendar_box.set_sensitive(0)
            self.quality_box.set_sensitive(0)
            self.dual_dated.set_sensitive(0)
            self.new_year.set_sensitive(0)

        self.text_entry = self.top.get_object("date_text_entry")
        self.text_entry.set_text(self.date.get_text())

        if self.date.get_slash():
            self.dual_dated.set_active(1)
            self.calendar_box.set_sensitive(0)
            self.calendar_box.set_active(Date.CAL_JULIAN)
        self.dual_dated.connect("toggled", self.switch_dual_dated)

        # Use the calendar from user preferences for new dates only.
        if self.date.is_empty():
            cal = config.get("preferences.calendar-format-input")
            self.calendar_box.set_active(cal)

        # The dialog is modal -- since dates don't have names, we don't
        # want to have several open dialogs, since then the user will
        # loose track of which is which. Much like opening files.

        self.validated_date = self.return_date = None

        for obj in self.top.get_objects():
            if obj != self.ok_button:
                for signal in ["changed", "value-changed"]:
                    try:
                        obj.connect_after(signal, self.revalidate)
                    except TypeError:
                        # some of them don't support the signal, ignore them...
                        pass
        self.revalidate()
        self.show()

        while True:
            response = self.window.run()
            LOG.debug("response: {0}".format(response))
            if response == Gtk.ResponseType.HELP:
                display_help(webpage=WIKI_HELP_PAGE, section=WIKI_HELP_SEC)
            elif response == Gtk.ResponseType.DELETE_EVENT:
                break
            else:
                if response == Gtk.ResponseType.OK:
                    # if the user pressed OK/enter while inside an edit field,
                    # e.g., the year,
                    # build_date_from_ui won't pick up the new text in the
                    # run of revalidate that allowed the OK!
                    if not self.revalidate():
                        continue
                    self.return_date = Date()
                    self.return_date.copy(self.validated_date)
                self.close()
                break

    def revalidate(self, obj=None):
        """
        If anything changed, revalidate the date and
        enable/disable the "OK" button based on the result.
        """
        (
            the_quality,
            the_modifier,
            the_calendar,
            the_value,
            the_text,
            the_newyear,
        ) = self.build_date_from_ui()
        LOG.debug("revalidate: {0} changed, value: {1}".format(obj, the_value))
        dat = Date(self.date)
        if not self.ok_button.get_sensitive():
            self.statusbar.pop(1)
        try:
            dat.set(
                quality=the_quality,
                modifier=the_modifier,
                calendar=the_calendar,
                value=the_value,
                text=the_text,
                newyear=the_newyear,
            )
            # didn't throw yet?
            self.validated_date = dat
            LOG.debug("validated_date set to: {0}".format(dat.__dict__))
            self.ok_button.set_sensitive(1)
            self.calendar_box.set_sensitive(1)
            return True
        except DateError as dummy_err:
            self.ok_button.set_sensitive(0)
            self.calendar_box.set_sensitive(0)
            curmode = MOD_TEXT[self.type_box.get_active()][1]
            txtmode = MOD_TEXT[-1][1]
            self.statusbar.push(
                1,
                _(
                    "Correct the date or switch from `{cur_mode}'" " to `{text_mode}'"
                ).format(cur_mode=curmode, text_mode=txtmode),
            )
            return False

    def build_menu_names(self, obj):
        """
        Define the menu entry for the :class:`.ManagedWindow`
        """
        return (_("Date selection"), None)

    def build_date_from_ui(self):
        """
        Collect information from the UI controls and return
        5-tuple of (quality,modifier,calendar,value,text)
        """
        # It is important to not set date based on these controls.
        # For example, changing the caledar makes the date inconsistent
        # until the callback of the calendar menu is finished.
        # We need to be able to use this function from that callback,
        # so here we just report on the state of all widgets, without
        # actually modifying the date yet.
        modifier = MOD_TEXT[self.type_box.get_active()][0]
        text = self.text_entry.get_text()

        if modifier == Date.MOD_TEXTONLY:
            return (
                Date.QUAL_NONE,
                Date.MOD_TEXTONLY,
                Date.CAL_GREGORIAN,
                Date.EMPTY,
                text,
                Date.NEWYEAR_JAN1,
            )

        quality = QUAL_TEXT[self.quality_box.get_active()][0]

        if modifier in (Date.MOD_RANGE, Date.MOD_SPAN):
            value = (
                self.start_day.get_value_as_int(),
                self.start_month_box.get_active(),
                self.start_year.get_value_as_int(),
                self.dual_dated.get_active(),
                self.stop_day.get_value_as_int(),
                self.stop_month_box.get_active(),
                self.stop_year.get_value_as_int(),
                self.dual_dated.get_active(),
            )
        else:
            value = (
                self.start_day.get_value_as_int(),
                self.start_month_box.get_active(),
                self.start_year.get_value_as_int(),
                self.dual_dated.get_active(),
            )
        calendar = self.calendar_box.get_active()
        newyear = Date.newyear_to_code(self.new_year.get_text())
        return (quality, modifier, calendar, value, text, newyear)

    def switch_type(self, obj):
        """
        Disable/enable various date controls depending on the date
        type selected via the menu.
        """

        the_modifier = MOD_TEXT[self.type_box.get_active()][0]

        # Disable/enable second date controls based on whether
        # the type allows compound dates
        if the_modifier in (Date.MOD_RANGE, Date.MOD_SPAN):
            stop_date_sensitivity = 1
        else:
            stop_date_sensitivity = 0
        self.stop_day.set_sensitive(stop_date_sensitivity)
        self.stop_month_box.set_sensitive(stop_date_sensitivity)
        self.stop_year.set_sensitive(stop_date_sensitivity)

        # Disable/enable the rest of the controls if the type is text-only.
        date_sensitivity = not the_modifier == Date.MOD_TEXTONLY
        self.start_day.set_sensitive(date_sensitivity)
        self.start_month_box.set_sensitive(date_sensitivity)
        self.start_year.set_sensitive(date_sensitivity)
        self.calendar_box.set_sensitive(date_sensitivity)
        self.quality_box.set_sensitive(date_sensitivity)
        self.dual_dated.set_sensitive(date_sensitivity)
        self.new_year.set_sensitive(date_sensitivity)

    def switch_dual_dated(self, obj):
        """
        Changed whether this is a dual dated year, or not.
        Dual dated years are represented in the Julian calendar
        so that the day/months don't changed in the Text representation.
        """
        if self.dual_dated.get_active():
            self.calendar_box.set_active(Date.CAL_JULIAN)
            self.calendar_box.set_sensitive(0)
        else:
            self.calendar_box.set_sensitive(1)

    def align_newyear_ui_with_calendar(self, cal):
        if calendar_has_fixed_newyear(cal):
            LOG.debug("new year disabled for cal {0}".format(cal))
            self.new_year.set_sensitive(0)
            self.new_year.set_text("")
        else:
            LOG.debug("new year enabled for cal {0}".format(cal))
            self.new_year.set_sensitive(1)

    def switch_calendar(self, obj):
        """
        Change month names and convert the date based on the calendar
        selected via the menu.
        """

        old_cal = self.date.get_calendar()
        new_cal = self.calendar_box.get_active()
        LOG.debug(
            ">>>switch_calendar: {0} changed, {1} -> {2}".format(obj, old_cal, new_cal)
        )

        self.align_newyear_ui_with_calendar(new_cal)

        (
            the_quality,
            the_modifier,
            dummy_the_calendar,
            the_value,
            the_text,
            the_newyear,
        ) = self.build_date_from_ui()
        try:
            self.date.set(
                quality=the_quality,
                modifier=the_modifier,
                calendar=old_cal,
                value=the_value,
                text=the_text,
                newyear=the_newyear,
            )
        except DateError:
            pass
        else:
            if not self.date.is_empty():
                self.date.convert_calendar(new_cal)

        self.start_month_box.get_model().clear()
        self.stop_month_box.get_model().clear()
        month_names = CAL_TO_MONTHS_NAMES[new_cal]
        for name in month_names:
            self.start_month_box.append_text(name)
            self.stop_month_box.append_text(name)

        self.start_day.set_value(self.date.get_day())
        self.start_month_box.set_active(self.date.get_month())
        self.start_year.set_value(self.date.get_year())
        self.stop_day.set_value(self.date.get_stop_day())
        self.stop_month_box.set_active(self.date.get_stop_month())
        self.stop_year.set_value(self.date.get_stop_year())
        LOG.debug(
            "<<<switch_calendar: {0} changed, {1} -> {2}".format(obj, old_cal, new_cal)
        )



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import List, Set, Tuple
from ....lib import Person
from ....db import Database
from ....types import PersonHandle


# -------------------------------------------------------------------------
#
# IsLessThanNthGenerationAncestorOf
#
# -------------------------------------------------------------------------
class IsLessThanNthGenerationAncestorOf(Rule):
    """Rule that checks for a person that is an ancestor of a specified person
    not more than N generations away"""

    labels = [_("ID:"), _("Number of generations:")]
    name = _("Ancestors of <person> not more than <N> generations away")
    category = _("Ancestral filters")
    description = _(
        "Matches people that are ancestors "
        "of a specified person not more than N generations away"
    )

    def prepare(self, db: Database, user):
        self.db = db
        self.selected_handles: Set[str] = set()
        person = db.get_person_from_gramps_id(self.list[0])
        if person:
            root_handle = person.handle
            if root_handle:
                self.init_ancestor_list(root_handle)

    def init_ancestor_list(self, root_handle: PersonHandle):
        queue: List[Tuple[PersonHandle, int]] = [
            (root_handle, 1)
        ]  # generation 1 is root
        while queue:
            handle, gen = queue.pop(0)
            if handle in self.selected_handles:
                # if we have been here before, skip
                continue
            self.selected_handles.add(handle)
            gen += 1
            if gen <= int(self.list[1]):
                p = self.db.get_person_from_handle(handle)
                fam_id = (
                    p.parent_family_list[0] if len(p.parent_family_list) > 0 else None
                )
                if fam_id:
                    fam = self.db.get_family_from_handle(fam_id)
                    if fam:
                        f_id = fam.father_handle
                        m_id = fam.mother_handle
                        # append to back of queue:
                        if f_id:
                            queue.append((f_id, gen))
                        if m_id:
                            queue.append((m_id, gen))

    def reset(self):
        self.selected_handles.clear()

    def apply_to_one(self, db: Database, person: Person) -> bool:
        return person.handle in self.selected_handles



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2008 Douglas S. Blank
# Copyright (C) 2004-2007 Donald N. Allingham
# Copyright (C) 2008      Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Vassilii Khachaturov
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"Export to CSV Spreadsheet."

# -------------------------------------------------------------------------
#
# Standard Python Modules
#
# -------------------------------------------------------------------------
import os
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
import csv
from io import StringIO
import codecs

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
import logging
from collections import abc

LOG = logging.getLogger(".ExportCSV")

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.config import config
from gramps.gen.lib import EventType, Person
from gramps.gen.lib.eventroletype import EventRoleType
from gramps.gui.plug.export import WriterOptionBox
from gramps.gen.utils.string import gender as gender_map
from gramps.gen.datehandler import get_date
from gramps.gen.display.place import displayer as _pd
from gramps.gui.glade import Glade
from gramps.gen.constfunc import win


# -------------------------------------------------------------------------
#
# The function that does the exporting
#
# -------------------------------------------------------------------------
def exportData(database, filename, user, option_box=None):
    gw = CSVWriter(database, filename, user, option_box)
    return gw.export_data()


# -------------------------------------------------------------------------
#
# Support Functions
#
# -------------------------------------------------------------------------
def sortable_string_representation(text):
    numeric = ""
    alpha = ""
    for s in text:
        if s.isdigit():
            numeric += s
        else:
            alpha += s
    return alpha + (("0" * 10) + numeric)[-10:]


def get_primary_event_ref_from_type(db, person, event_name):
    """
    >>> get_primary_event_ref_from_type(db, Person(), "Baptism"):
    """
    for ref in person.event_ref_list:
        if ref.get_role() == EventRoleType.PRIMARY:
            event = db.get_event_from_handle(ref.ref)
            if event and event.type.is_type(event_name):
                return ref
    return None


def get_primary_source_title(db, obj):
    for citation_handle in obj.get_citation_list():
        citation = db.get_citation_from_handle(citation_handle)
        source = db.get_source_from_handle(citation.get_reference_handle())
        if source:
            return source.get_title()
    return ""


# -------------------------------------------------------------------------
#
# CSVWriter Options
#
# -------------------------------------------------------------------------
class CSVWriterOptionBox(WriterOptionBox):
    """
    Create a VBox with the option widgets and define methods to retrieve
    the options.

    """

    def __init__(self, person, dbstate, uistate, track=[], window=None):
        WriterOptionBox.__init__(
            self, person, dbstate, uistate, track=track, window=window
        )
        ## TODO: add place filter selection
        self.include_individuals = 1
        self.include_marriages = 1
        self.include_children = 1
        self.include_places = 1
        self.translate_headers = 1
        self.include_individuals_check = None
        self.include_marriages_check = None
        self.include_children_check = None
        self.include_places_check = None
        self.translate_headers_check = None

    def get_option_box(self):
        from gi.repository import Gtk

        option_box = WriterOptionBox.get_option_box(self)

        self.include_individuals_check = Gtk.CheckButton(label=_("Include people"))
        self.include_marriages_check = Gtk.CheckButton(label=_("Include marriages"))
        self.include_children_check = Gtk.CheckButton(label=_("Include children"))
        self.include_places_check = Gtk.CheckButton(label=_("Include places"))
        self.translate_headers_check = Gtk.CheckButton(label=_("Translate headers"))

        self.include_individuals_check.set_active(1)
        self.include_marriages_check.set_active(1)
        self.include_children_check.set_active(1)
        self.include_places_check.set_active(1)
        self.translate_headers_check.set_active(1)

        option_box.pack_start(self.include_individuals_check, False, True, 0)
        option_box.pack_start(self.include_marriages_check, False, True, 0)
        option_box.pack_start(self.include_children_check, False, True, 0)
        option_box.pack_start(self.include_places_check, False, True, 0)
        option_box.pack_start(self.translate_headers_check, False, True, 0)

        return option_box

    def parse_options(self):
        WriterOptionBox.parse_options(self)
        if self.include_individuals_check:
            self.include_individuals = self.include_individuals_check.get_active()
            self.include_marriages = self.include_marriages_check.get_active()
            self.include_children = self.include_children_check.get_active()
            self.include_places = self.include_places_check.get_active()
            self.translate_headers = self.translate_headers_check.get_active()


# -------------------------------------------------------------------------
#
# CSVWriter class
#
# -------------------------------------------------------------------------
class CSVWriter:
    def __init__(self, database, filename, user, option_box=None):
        self.db = database
        self.option_box = option_box
        self.filename = filename
        self.user = user
        if isinstance(self.user.callback, abc.Callable):  # is really callable
            self.update = self.update_real
        else:
            self.update = self.update_empty

        self.plist = {}
        self.flist = {}
        self.place_list = {}

        self.persons_details_done = []
        self.persons_notes_done = []
        self.person_ids = {}

        if not option_box:
            self.include_individuals = 1
            self.include_marriages = 1
            self.include_children = 1
            self.include_places = 1
            self.translate_headers = 1
        else:
            self.option_box.parse_options()
            self.db = option_box.get_filtered_database(self.db)

            self.include_individuals = self.option_box.include_individuals
            self.include_marriages = self.option_box.include_marriages
            self.include_children = self.option_box.include_children
            self.include_places = self.option_box.include_places
            self.translate_headers = self.option_box.translate_headers

        self.plist = [x for x in self.db.iter_person_handles()]

        # make place list so that dependencies are first:
        self.place_list = []
        place_list = sorted([x for x in self.db.iter_place_handles()])
        while place_list:
            handle = place_list[0]
            place = self.db.get_place_from_handle(handle)
            if place:
                if all([(x.ref in self.place_list) for x in place.placeref_list]):
                    self.place_list.append(place_list.pop(0))
                else:  # put at the back of the line:
                    place_list.append(place_list.pop(0))
            else:
                place_list.pop(0)
        # get the families for which these people are spouses:
        self.flist = {}
        for key in self.plist:
            p = self.db.get_person_from_handle(key)
            if p:
                for family_handle in p.get_family_handle_list():
                    self.flist[family_handle] = 1
        # now add the families for which these people are a child:
        for family_handle in self.db.iter_family_handles():
            family = self.db.get_family_from_handle(family_handle)
            if family:
                for child_ref in family.get_child_ref_list():
                    if child_ref:
                        child_handle = child_ref.ref
                        if child_handle in self.plist:
                            self.flist[family_handle] = 1

    def update_empty(self):
        pass

    def update_real(self):
        self.count += 1
        newval = int(100 * self.count / self.total)
        if newval != self.oldval:
            self.user.callback(newval)
            self.oldval = newval

    def writeln(self):
        self.g.writerow([])

    def write_csv(self, *items):
        self.g.writerow(items)

    def export_data(self):
        self.dirname = os.path.dirname(self.filename)
        try:
            self.fp = open(
                self.filename,
                "w",
                encoding="utf_8_sig" if win() else "utf_8",
                newline="",
            )
            my_dialect = config.get("csv.dialect")
            my_delimiter = config.get("csv.delimiter")
            if my_dialect == _("Custom"):
                self.g = csv.writer(self.fp, delimiter=my_delimiter)
            else:
                self.g = csv.writer(self.fp, dialect=my_dialect)
        except IOError as msg:
            msg2 = _("Could not create %s") % self.filename
            self.user.notify_error(msg2, str(msg))
            return False
        except:
            self.user.notify_error(_("Could not create %s") % self.filename)
            return False
        ######################### initialize progress bar
        self.count = 0
        self.total = 0
        self.oldval = 0
        if self.include_individuals:
            self.total += len(self.plist)
        if self.include_marriages:
            self.total += len(self.flist)
        if self.include_children:
            self.total += len(self.flist)
        if self.include_places:
            self.total += len(self.place_list)
        ########################
        LOG.debug("Possible people to export: %s", len(self.plist))
        LOG.debug("Possible families to export: %s", len(self.flist))
        LOG.debug("Possible places to export: %s", len(self.place_list))
        ###########################
        if self.include_places:
            if self.translate_headers:
                self.write_csv(
                    _("Place"),
                    _("Title"),
                    _("Name"),
                    _("Type"),
                    _("Latitude"),
                    _("Longitude"),
                    _("Code"),
                    _("Enclosed_by"),
                    _("Date"),
                )
            else:
                self.write_csv(
                    "Place",
                    "Title",
                    "Name",
                    "Type",
                    "Latitude",
                    "Longitude",
                    "Code",
                    "Enclosed_by",
                    "Date",
                )
            for key in self.place_list:
                place = self.db.get_place_from_handle(key)
                if place:
                    place_id = place.gramps_id
                    place_title = place.title
                    place_name = place.name.value
                    place_type = str(place.place_type)
                    place_latitude = place.lat
                    place_longitude = place.long
                    place_code = place.code
                    if place.placeref_list:
                        for placeref in place.placeref_list:
                            placeref_obj = self.db.get_place_from_handle(placeref.ref)
                            placeref_date = ""
                            if not placeref.date.is_empty():
                                placeref_date = placeref.date
                            placeref_id = ""
                            if placeref_obj:
                                placeref_id = "[%s]" % placeref_obj.gramps_id
                            self.write_csv(
                                "[%s]" % place_id,
                                place_title,
                                place_name,
                                place_type,
                                place_latitude,
                                place_longitude,
                                place_code,
                                placeref_id,
                                placeref_date,
                            )
                    else:
                        self.write_csv(
                            "[%s]" % place_id,
                            place_title,
                            place_name,
                            place_type,
                            place_latitude,
                            place_longitude,
                            place_code,
                            "",
                            "",
                        )
                self.update()
            self.writeln()
        ########################### sort:
        sortorder = []
        dropped_surnames = set()
        for key in self.plist:
            person = self.db.get_person_from_handle(key)
            if person:
                primary_name = person.get_primary_name()
                first_name = primary_name.get_first_name()
                surname_obj = primary_name.get_primary_surname()
                surname = surname_obj.get_surname()

                # See bug #6955
                nonprimary_surnames = set(primary_name.get_surname_list())
                nonprimary_surnames.remove(surname_obj)
                dropped_surnames.update(nonprimary_surnames)

                sortorder.append((surname, first_name, key))
        if dropped_surnames:
            LOG.warning(
                _(
                    "CSV export doesn't support non-primary surnames, "
                    "{count} dropped"
                ).format(count=len(dropped_surnames))
            )
            LOG.debug(
                "Dropped surnames: "
                + ", ".join(
                    [
                        (
                            "%s %s %s"
                            % (
                                surname.get_prefix(),
                                surname.get_surname(),
                                surname.get_connector(),
                            )
                        ).strip()
                        for surname in dropped_surnames
                    ]
                )
            )
        sortorder.sort()  # will sort on tuples
        plist = [data[2] for data in sortorder]
        ###########################
        if self.include_individuals:
            if self.translate_headers:
                self.write_csv(
                    _("Person"),
                    _("Surname"),
                    _("Given"),
                    _("Call"),
                    _("Nickname"),
                    _("Suffix"),
                    _("Prefix"),
                    _("Title", "Person"),
                    _("Gender"),
                    _("Birth date"),
                    _("Birth place"),
                    _("Birth source"),
                    _("Baptism date"),
                    _("Baptism place"),
                    _("Baptism source"),
                    _("Death date"),
                    _("Death place"),
                    _("Death source"),
                    _("Burial date"),
                    _("Burial place"),
                    _("Burial source"),
                    _("Note"),
                )
            else:
                self.write_csv(
                    "Person",
                    "Surname",
                    "Given",
                    "Call",
                    "Nickname",
                    "Suffix",
                    "Prefix",
                    "Title",
                    "Gender",
                    "Birth date",
                    "Birth place",
                    "Birth source",
                    "Baptism date",
                    "Baptism place",
                    "Baptism source",
                    "Death date",
                    "Death place",
                    "Death source",
                    "Burial date",
                    "Burial place",
                    "Burial source",
                    "Note",
                )
            for key in plist:
                person = self.db.get_person_from_handle(key)
                if person:
                    primary_name = person.get_primary_name()
                    first_name = primary_name.get_first_name()
                    surname_obj = primary_name.get_primary_surname()
                    surname = surname_obj.get_surname()
                    prefix = surname_obj.get_prefix()
                    suffix = primary_name.get_suffix()
                    title = primary_name.get_title()
                    grampsid = person.get_gramps_id()
                    grampsid_ref = ""
                    if grampsid != "":
                        grampsid_ref = "[" + grampsid + "]"
                    note = ""  # don't export notes
                    callname = primary_name.get_call_name()
                    nickname = primary_name.get_nick_name()
                    gender = person.get_gender()
                    if gender == Person.MALE:
                        gender = gender_map[Person.MALE]
                    elif gender == Person.FEMALE:
                        gender = gender_map[Person.FEMALE]
                    elif gender == Person.OTHER:
                        gender = gender_map[Person.OTHER]
                    else:
                        gender = gender_map[Person.UNKNOWN]
                    # Birth:
                    birthdate = ""
                    birthplace = ""
                    birthsource = ""
                    birth_ref = person.get_birth_ref()
                    if birth_ref:
                        birth = self.db.get_event_from_handle(birth_ref.ref)
                        if birth:
                            birthdate = self.format_date(birth)
                            birthplace = self.format_place(birth)
                            birthsource = get_primary_source_title(self.db, birth)
                    # Baptism:
                    baptismdate = ""
                    baptismplace = ""
                    baptismsource = ""
                    baptism_ref = get_primary_event_ref_from_type(
                        self.db, person, "Baptism"
                    )
                    if baptism_ref:
                        baptism = self.db.get_event_from_handle(baptism_ref.ref)
                        if baptism:
                            baptismdate = self.format_date(baptism)
                            baptismplace = self.format_place(baptism)
                            baptismsource = get_primary_source_title(self.db, baptism)
                    # Death:
                    deathdate = ""
                    deathplace = ""
                    deathsource = ""
                    death_ref = person.get_death_ref()
                    if death_ref:
                        death = self.db.get_event_from_handle(death_ref.ref)
                        if death:
                            deathdate = self.format_date(death)
                            deathplace = self.format_place(death)
                            deathsource = get_primary_source_title(self.db, death)
                    # Burial:
                    burialdate = ""
                    burialplace = ""
                    burialsource = ""
                    burial_ref = get_primary_event_ref_from_type(
                        self.db, person, "Burial"
                    )
                    if burial_ref:
                        burial = self.db.get_event_from_handle(burial_ref.ref)
                        if burial:
                            burialdate = self.format_date(burial)
                            burialplace = self.format_place(burial)
                            burialsource = get_primary_source_title(self.db, burial)
                    # Write it out:
                    self.write_csv(
                        grampsid_ref,
                        surname,
                        first_name,
                        callname,
                        nickname,
                        suffix,
                        prefix,
                        title,
                        gender,
                        birthdate,
                        birthplace,
                        birthsource,
                        baptismdate,
                        baptismplace,
                        baptismsource,
                        deathdate,
                        deathplace,
                        deathsource,
                        burialdate,
                        burialplace,
                        burialsource,
                        note,
                    )
                self.update()
            self.writeln()
        ########################### sort:
        sortorder = []
        for key in self.flist:
            family = self.db.get_family_from_handle(key)
            if family:
                marriage_id = family.get_gramps_id()
                sortorder.append((sortable_string_representation(marriage_id), key))
        sortorder.sort()  # will sort on tuples
        flist = [data[1] for data in sortorder]
        ###########################
        if self.include_marriages:
            if self.translate_headers:
                self.write_csv(
                    _("Marriage"),
                    _("Husband"),
                    _("Wife"),
                    _("Date"),
                    _("Place"),
                    _("Source"),
                    _("Note"),
                )
            else:
                self.write_csv(
                    "Marriage", "Husband", "Wife", "Date", "Place", "Source", "Note"
                )
            for key in flist:
                family = self.db.get_family_from_handle(key)
                if family:
                    marriage_id = family.get_gramps_id()
                    if marriage_id != "":
                        marriage_id = "[" + marriage_id + "]"
                    mother_id = ""
                    father_id = ""
                    father_handle = family.get_father_handle()
                    if father_handle:
                        father = self.db.get_person_from_handle(father_handle)
                        father_id = father.get_gramps_id()
                        if father_id != "":
                            father_id = "[" + father_id + "]"
                    mother_handle = family.get_mother_handle()
                    if mother_handle:
                        mother = self.db.get_person_from_handle(mother_handle)
                        mother_id = mother.get_gramps_id()
                        if mother_id != "":
                            mother_id = "[" + mother_id + "]"
                    # get mdate, mplace
                    mdate, mplace, source = "", "", ""
                    event_ref_list = family.get_event_ref_list()
                    for event_ref in event_ref_list:
                        event = self.db.get_event_from_handle(event_ref.ref)
                        if event.get_type() == EventType.MARRIAGE:
                            mdate = self.format_date(event)
                            mplace = self.format_place(event)
                            source = get_primary_source_title(self.db, event)
                    note = ""
                    self.write_csv(
                        marriage_id, father_id, mother_id, mdate, mplace, source, note
                    )
                self.update()
            self.writeln()
        if self.include_children:
            if self.translate_headers:
                self.write_csv(_("Family"), _("Child"))
            else:
                self.write_csv("Family", "Child")
            for key in flist:
                family = self.db.get_family_from_handle(key)
                if family:
                    family_id = family.get_gramps_id()
                    if family_id != "":
                        family_id = "[" + family_id + "]"
                    for child_ref in family.get_child_ref_list():
                        child_handle = child_ref.ref
                        child = self.db.get_person_from_handle(child_handle)
                        grampsid = child.get_gramps_id()
                        grampsid_ref = ""
                        if grampsid != "":
                            grampsid_ref = "[" + grampsid + "]"
                        self.write_csv(family_id, grampsid_ref)
                self.update()
            self.writeln()
        self.fp.close()
        return True

    def format_date(self, date):
        return get_date(date)

    def format_place(self, event):
        """
        If places are included in the export return a link, else return a
        formatted place for the given event.
        """
        if self.include_places:
            place_handle = event.get_place_handle()
            if place_handle:
                place = self.db.get_place_from_handle(place_handle)
                if place:
                    return "[%s]" % place.get_gramps_id()
            return ""
        else:
            return _pd.display_event(self.db, event)



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Catalan Version, 2008

"""
Catalan-specific classes for parsing and displaying dates.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import re

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib.date import Date
from ._dateparser import DateParser
from ._datedisplay import DateDisplay
from ._datehandler import register_datehandler


# -------------------------------------------------------------------------
#
# Catalan parser
#
# -------------------------------------------------------------------------
class DateParserCA(DateParser):
    modifier_to_int = {
        "abans de": Date.MOD_BEFORE,
        "abans": Date.MOD_BEFORE,
        "ab.": Date.MOD_BEFORE,
        "després de": Date.MOD_AFTER,
        "després": Date.MOD_AFTER,
        "desp.": Date.MOD_AFTER,
        "desp": Date.MOD_AFTER,
        "aprox.": Date.MOD_ABOUT,
        "aprox": Date.MOD_ABOUT,
        "circa": Date.MOD_ABOUT,
        "ca.": Date.MOD_ABOUT,
        "ca": Date.MOD_ABOUT,
        "c.": Date.MOD_ABOUT,
        "cap a": Date.MOD_ABOUT,
        "al voltant": Date.MOD_ABOUT,
        "al voltant de": Date.MOD_ABOUT,
        "des de": Date.MOD_FROM,
        "fins": Date.MOD_TO,
    }

    calendar_to_int = {
        "gregorià": Date.CAL_GREGORIAN,
        "g": Date.CAL_GREGORIAN,
        "julià": Date.CAL_JULIAN,
        "j": Date.CAL_JULIAN,
        "hebreu": Date.CAL_HEBREW,
        "h": Date.CAL_HEBREW,
        "islàmic": Date.CAL_ISLAMIC,
        "i": Date.CAL_ISLAMIC,
        "revolucionari": Date.CAL_FRENCH,
        "r": Date.CAL_FRENCH,
        "persa": Date.CAL_PERSIAN,
        "p": Date.CAL_PERSIAN,
        "swedish": Date.CAL_SWEDISH,
        "s": Date.CAL_SWEDISH,
    }

    quality_to_int = {
        "estimat": Date.QUAL_ESTIMATED,
        "est.": Date.QUAL_ESTIMATED,
        "est": Date.QUAL_ESTIMATED,
        "calc.": Date.QUAL_CALCULATED,
        "calc": Date.QUAL_CALCULATED,
        "calculat": Date.QUAL_CALCULATED,
    }

    def init_strings(self):
        DateParser.init_strings(self)
        _span_1 = ["des de"]
        _span_2 = ["fins a"]
        _range_1 = ["entre", r"ent\.", "ent"]
        _range_2 = ["i"]
        self._span = re.compile(
            r"(%s)\s+(?P<start>.+)\s+(%s)\s+(?P<stop>.+)"
            % ("|".join(_span_1), "|".join(_span_2)),
            re.IGNORECASE,
        )
        self._range = re.compile(
            r"(%s)\s+(?P<start>.+)\s+(%s)\s+(?P<stop>.+)"
            % ("|".join(_range_1), "|".join(_range_2)),
            re.IGNORECASE,
        )


# -------------------------------------------------------------------------
#
# Catalan display
#
# -------------------------------------------------------------------------
class DateDisplayCA(DateDisplay):
    """
    Catalan language date display class.
    """

    long_months = (
        "",
        "Gener",
        "Febrer",
        "Març",
        "Abril",
        "Maig",
        "Juny",
        "Juliol",
        "Agost",
        "Setembre",
        "Octubre",
        "Novembre",
        "Desembre",
    )

    short_months = (
        "",
        "Gen",
        "Feb",
        "Mar",
        "Abr",
        "Mai",
        "Jun",
        "Jul",
        "Ago",
        "Set",
        "Oct",
        "Nov",
        "Des",
    )

    calendar = ("", "Julià", "Hebreu", "Revolucionari", "Persa", "Islàmic", "Suec")

    _mod_str = ("", "abans de ", "després de ", "cap a ", "", "", "")

    _qual_str = ("", "estimat ", "calculat ")

    french = (
        "",
        "Vendemiari",
        "Brumari",
        "Frimari",
        "Nivós",
        "Pluviós",
        "Ventós",
        "Germinal",
        "Floreal",
        "Pradial",
        "Messidor",
        "Termidor",
        "Fructidor",
        "Extra",
    )

    formats = (
        "AAAA-MM-DD (ISO)",
        "Numèrica",
        "Mes Dia, Any",
        "MES Dia, Any",
        "Dia Mes, Any",
        "Dia MES, Any",
    )
    # this must agree with DateDisplayEn's "formats" definition
    # (since no locale-specific _display_gregorian exists, here)

    def display(self, date):
        """
        Return a text string representing the date.
        """
        mod = date.get_modifier()
        cal = date.get_calendar()
        qual = date.get_quality()
        start = date.get_start_date()
        newyear = date.get_new_year()

        qual_str = self._qual_str[qual]

        if mod == Date.MOD_TEXTONLY:
            return date.get_text()
        elif start == Date.EMPTY:
            return ""
        elif mod == Date.MOD_SPAN:
            d1 = self.display_cal[cal](start)
            d2 = self.display_cal[cal](date.get_stop_date())
            scal = self.format_extras(cal, newyear)
            return "%s%s %s %s %s%s" % (qual_str, "des de", d1, "fins a", d2, scal)
        elif mod == Date.MOD_RANGE:
            d1 = self.display_cal[cal](start)
            d2 = self.display_cal[cal](date.get_stop_date())
            scal = self.format_extras(cal, newyear)
            return "%s%s %s %s %s%s" % (qual_str, "entre", d1, "i", d2, scal)
        else:
            text = self.display_cal[date.get_calendar()](start)
            scal = self.format_extras(cal, newyear)
            return "%s%s%s%s" % (qual_str, self._mod_str[mod], text, scal)


# -------------------------------------------------------------------------
#
# Register classes
#
# -------------------------------------------------------------------------
register_datehandler(
    ("ca_ES", "ca", "català", "Catalan", "ca_FR", "ca_AD", "ca_IT", ("%d/%m/%Y",)),
    DateParserCA,
    DateDisplayCA,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2009  B. Malengier
# Copyright (C) 2009  Swoon on bug tracker
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
File tests.
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import os
import tempfile
import unittest

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ...const import USER_HOME, USER_PICTURES, USER_PLUGINS, VERSION
from ...utils.file import media_path
from ...db.utils import make_database


# -------------------------------------------------------------------------
#
# FileTest class
#
# -------------------------------------------------------------------------
class FileTest(unittest.TestCase):
    """
    File tests.
    """

    def test_mediapath(self):
        """
        Test media path variables.
        """
        with tempfile.TemporaryDirectory() as tmpdirname:
            # Create database
            db = make_database("sqlite")
            path = os.path.join(tmpdirname, "utils_file_test")
            os.makedirs(path)
            db.load(path)

            # Test without db.mediapath set
            self.assertEqual(
                media_path(db),
                os.path.normcase(os.path.normpath(os.path.abspath(USER_PICTURES))),
            )

            # Test with absolute db.mediapath
            db.set_mediapath(os.path.abspath(USER_HOME) + "/test_abs")
            self.assertEqual(
                media_path(db),
                os.path.normcase(
                    os.path.normpath(os.path.abspath(USER_HOME + "/test_abs"))
                ),
            )

            # Test with relative db.mediapath
            db.set_mediapath("test_rel")
            self.assertEqual(
                media_path(db),
                os.path.normcase(
                    os.path.normpath(
                        os.path.abspath(tmpdirname + "/utils_file_test/test_rel")
                    )
                ),
            )

            # Test with environment variable
            db.set_mediapath("/test/{VERSION}/test_var")
            self.assertEqual(
                media_path(db),
                os.path.normcase(
                    os.path.normpath(os.path.abspath("/test/" + VERSION + "/test_var"))
                ),
            )
            db.set_mediapath("{USER_PLUGINS}/test_var")
            self.assertEqual(
                media_path(db),
                os.path.normcase(
                    os.path.normpath(os.path.abspath(USER_PLUGINS + "/test_var"))
                ),
            )
            db.set_mediapath("{VERSION}/test_var")
            self.assertEqual(
                media_path(db),
                os.path.normcase(
                    os.path.normpath(
                        os.path.abspath(
                            tmpdirname + "/utils_file_test/" + VERSION + "/test_var"
                        )
                    )
                ),
            )

            # Test with $GRAMPSHOME environment variable not set
            old_env = os.environ.copy()
            if "GRAMPSHOME" in os.environ:
                del os.environ["GRAMPSHOME"]
            db.set_mediapath("{GRAMPSHOME}/test_var")
            self.assertEqual(
                media_path(db),
                os.path.normcase(
                    os.path.normpath(os.path.abspath(USER_HOME + "/test_var"))
                ),
            )

            # Test with $GRAMPSHOME environment variable set
            os.environ["GRAMPSHOME"] = "/this/is/a/test"
            db.set_mediapath("{GRAMPSHOME}/test_var")
            self.assertEqual(
                media_path(db),
                os.path.normcase(
                    os.path.normpath(os.path.abspath("/this/is/a/test/test_var"))
                ),
            )

            # Restore environment
            os.environ = old_env


# -------------------------------------------------------------------------
#
# main
#
# -------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2024       Doug Blank
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Mixin for DbDir to enable find_from_handle and check_from_handle methods.
"""

# ------------------------------------------------------------------------------
#
# Gramps Modules
#
# ------------------------------------------------------------------------------
from gramps.gen.lib import (
    Person,
    Family,
    Event,
    Place,
    Source,
    Media,
    Repository,
    Note,
    Tag,
)
from gramps.gen.lib.json_utils import data_to_object


# ------------------------------------------------------------------------------
#
# DbMixin class
#
# ------------------------------------------------------------------------------
class DbMixin:
    """
    DbMixin -- a collection of methods to be added to the main
    gramps database class for use with import functions.  To enable these
    functions, add the following code to your module:

        if DbMixin not in database.__class__.__bases__:
        database.__class__.__bases__ = (DbMixin,) +  \
                                        database.__class__.__bases__

    where "database" is the object name of your instance of the gramps
    database.
    """

    def __find_primary_from_handle(
        self, handle, transaction, class_type, get_raw_obj_data, add_func
    ):
        """
        Find a primary object of class_type in the database from the passed
        handle.

        If no object exists, a new object is added to the database.

        @return: Returns a tuple, first the object, second a bool which is True
                 if the object is new
        @rtype: tuple
        """
        obj = class_type()
        handle = str(handle)
        new = True
        raw = get_raw_obj_data(handle)
        if raw is not None:
            obj = data_to_object(raw)
            # references create object with id None before object is really made
            if obj.gramps_id is not None:
                new = False
        else:
            obj.set_handle(handle)
            add_func(obj, transaction)
        return obj, new

    def __find_table_from_handle(
        self, handle, transaction, class_type, get_raw_obj_data, add_func
    ):
        """
        Find a table object of class_type in the database from the passed
        handle.

        If no object exists, a new object is added to the database.

        @return: Returns a tuple, first the object, second a bool which is True
                 if the object is new
        @rtype: tuple
        """
        obj = class_type()
        handle = str(handle)
        raw = get_raw_obj_data(handle)
        if raw is not None:
            obj = data_to_object(raw)
            return obj, False
        else:
            obj.set_handle(handle)
            add_func(obj, transaction)
            return obj, True

    def __check_primary_from_handle(
        self, handle, transaction, class_type, has_handle_func, add_func, set_gid
    ):
        """
        Check whether a primary object of class_type with the passed handle
        exists in the database.

        If no such object exists, a new object is added to the database.
        If set_gid then a new gramps_id is created, if not, None is used.
        """
        handle = str(handle)
        if not has_handle_func(handle):
            obj = class_type()
            obj.set_handle(handle)
            add_func(obj, transaction, set_gid=set_gid)

    def __check_table_from_handle(
        self, handle, transaction, class_type, has_handle_func, add_func
    ):
        """
        Check whether a table object of class_type with the passed handle exists
        in the database.

        If no such object exists, a new object is added to the database.
        """
        handle = str(handle)
        if not has_handle_func(handle):
            obj = class_type()
            obj.set_handle(handle)
            add_func(obj, transaction)

    def find_person_from_handle(self, handle, transaction):
        """
        Find a Person in the database from the passed handle.

        If no such Person exists, a new Person is added to the database.

        @return: Returns a tuple, first the object, second a bool which is True
                 if the object is new
        @rtype: tuple
        """
        return self.__find_primary_from_handle(
            handle, transaction, Person, self.get_raw_person_data, self.add_person
        )

    def find_source_from_handle(self, handle, transaction):
        """
        Find a Source in the database from the passed handle.

        If no such Source exists, a new Source is added to the database.

        @return: Returns a tuple, first the object, second a bool which is True
                 if the object is new
        @rtype: tuple
        """
        return self.__find_primary_from_handle(
            handle, transaction, Source, self.get_raw_source_data, self.add_source
        )

    def find_event_from_handle(self, handle, transaction):
        """
        Find a Event in the database from the passed handle.

        If no such Event exists, a new Event is added to the database.

        @return: Returns a tuple, first the object, second a bool which is True
                 if the object is new
        @rtype: tuple
        """
        return self.__find_primary_from_handle(
            handle, transaction, Event, self.get_raw_event_data, self.add_event
        )

    def find_object_from_handle(self, handle, transaction):
        """
        Find a Media in the database from the passed handle.

        If no such Media exists, a new Object is added to the database.

        @return: Returns a tuple, first the object, second a bool which is True
                 if the object is new
        @rtype: tuple
        """
        return self.__find_primary_from_handle(
            handle, transaction, Media, self.get_raw_object_data, self.add_object
        )

    def find_place_from_handle(self, handle, transaction):
        """
        Find a Place in the database from the passed handle.

        If no such Place exists, a new Place is added to the database.

        @return: Returns a tuple, first the object, second a bool which is True
                 if the object is new
        @rtype: tuple
        """
        return self.__find_primary_from_handle(
            handle, transaction, Place, self.get_raw_place_data, self.add_place
        )

    def find_family_from_handle(self, handle, transaction):
        """
        Find a Family in the database from the passed handle.

        If no such Family exists, a new Family is added to the database.

        @return: Returns a tuple, first the object, second a bool which is True
                 if the object is new
        @rtype: tuple
        """
        return self.__find_primary_from_handle(
            handle, transaction, Family, self.get_raw_family_data, self.add_family
        )

    def find_repository_from_handle(self, handle, transaction):
        """
        Find a Repository in the database from the passed handle.

        If no such Repository exists, a new Repository is added to the database.

        @return: Returns a tuple, first the object, second a bool which is True
                 if the object is new
        @rtype: tuple
        """
        return self.__find_primary_from_handle(
            handle,
            transaction,
            Repository,
            self.get_raw_repository_data,
            self.add_repository,
        )

    def find_note_from_handle(self, handle, transaction):
        """
        Find a Note in the database from the passed handle.

        If no such Note exists, a new Note is added to the database.

        @return: Returns a tuple, first the object, second a bool which is True
                 if the object is new
        @rtype: tuple
        """
        return self.__find_primary_from_handle(
            handle, transaction, Note, self.get_raw_note_data, self.add_note
        )

    def find_tag_from_handle(self, handle, transaction):
        """
        Find a Tag in the database from the passed handle.

        If no such Tag exists, a new Tag is added to the database.

        @return: Returns a tuple, first the object, second a bool which is True
                 if the object is new
        @rtype: tuple
        """
        return self.__find_table_from_handle(
            handle, transaction, Tag, self.get_raw_tag_data, self.add_tag
        )

    def check_person_from_handle(self, handle, transaction, set_gid=True):
        """
        Check whether a Person with the passed handle exists in the database.

        If no such Person exists, a new Person is added to the database.
        If set_gid then a new gramps_id is created, if not, None is used.
        """
        self.__check_primary_from_handle(
            handle,
            transaction,
            Person,
            self.has_person_handle,
            self.add_person,
            set_gid=set_gid,
        )

    def check_source_from_handle(self, handle, transaction, set_gid=True):
        """
        Check whether a Source with the passed handle exists in the database.

        If no such Source exists, a new Source is added to the database.
        If set_gid then a new gramps_id is created, if not, None is used.
        """
        self.__check_primary_from_handle(
            handle,
            transaction,
            Source,
            self.has_source_handle,
            self.add_source,
            set_gid=set_gid,
        )

    def check_event_from_handle(self, handle, transaction, set_gid=True):
        """
        Check whether an Event with the passed handle exists in the database.

        If no such Event exists, a new Event is added to the database.
        If set_gid then a new gramps_id is created, if not, None is used.
        """
        self.__check_primary_from_handle(
            handle,
            transaction,
            Event,
            self.has_event_handle,
            self.add_event,
            set_gid=set_gid,
        )

    def check_object_from_handle(self, handle, transaction, set_gid=True):
        """
        Check whether a Media with the passed handle exists in the
        database.

        If no such Media exists, a new Object is added to the database.
        If set_gid then a new gramps_id is created, if not, None is used.
        """

        self.__check_primary_from_handle(
            handle,
            transaction,
            Media,
            self.has_object_handle,
            self.add_object,
            set_gid=set_gid,
        )

    def check_place_from_handle(self, handle, transaction, set_gid=True):
        """
        Check whether a Place with the passed handle exists in the database.

        If no such Place exists, a new Place is added to the database.
        If set_gid then a new gramps_id is created, if not, None is used.
        """
        self.__check_primary_from_handle(
            handle,
            transaction,
            Place,
            self.has_place_handle,
            self.add_place,
            set_gid=set_gid,
        )

    def check_family_from_handle(self, handle, transaction, set_gid=True):
        """
        Check whether a Family with the passed handle exists in the database.

        If no such Family exists, a new Family is added to the database.
        If set_gid then a new gramps_id is created, if not, None is used.
        """
        self.__check_primary_from_handle(
            handle,
            transaction,
            Family,
            self.has_family_handle,
            self.add_family,
            set_gid=set_gid,
        )

    def check_repository_from_handle(self, handle, transaction, set_gid=True):
        """
        Check whether a Repository with the passed handle exists in the
        database.

        If no such Repository exists, a new Repository is added to the database.
        If set_gid then a new gramps_id is created, if not, None is used.
        """
        self.__check_primary_from_handle(
            handle,
            transaction,
            Repository,
            self.has_repository_handle,
            self.add_repository,
            set_gid=set_gid,
        )

    def check_note_from_handle(self, handle, transaction, set_gid=True):
        """
        Check whether a Note with the passed handle exists in the database.

        If no such Note exists, a new Note is added to the database.
        If set_gid then a new gramps_id is created, if not, None is used.
        """
        self.__check_primary_from_handle(
            handle,
            transaction,
            Note,
            self.has_note_handle,
            self.add_note,
            set_gid=set_gid,
        )

    def check_tag_from_handle(self, handle, transaction):
        """
        Check whether a Tag with the passed handle exists in the database.

        If no such Tag exists, a new Tag is added to the database.
        """
        self.__check_table_from_handle(
            handle, transaction, Tag, self.has_tag_handle, self.add_tag
        )



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2005  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2009-2010  Andrew I Baznikin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Written by Alex Roitman, largely based on relationship.py by Don Allingham.
"""
Russian-specific definitions of relationships
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

from gramps.gen.lib import Person
import gramps.gen.relationship

# -------------------------------------------------------------------------

_parents_level = [
    "",
    "родители",
    "дедушки/бабушки",
    "прадедушки/прабабушки",
    "прапрадедушки/прапрабабушки (5 поколение)",
    "прапрапрадедушки/прапрапрабабушки (6 поколение)",
    "прапрапрапрадедушки/прапрапрапрабабушки (7 поколение)",
    "прапрапрапрапрадедушки/прапрапрапрапрабабушки (8 поколение)",
]

_male_cousin_level = [
    "",
    "двоюродный",
    "троюродный",
    "четвероюродный",
    "пятиюродный",
    "шестиюродный",
    "семиюродный",
    "восьмиюродный",
    "девятиюродный",
    "десятиюродный",
    "одиннацатиюродный",
    "двенадцатиюродный",
    "тринадцатиюродный",
    "четырнадцатиюродный",
    "пятнадцатиюродный",
    "шестнадцатиюродный",
    "семнадцатиюродный",
    "восемнадцатиюродный",
    "девятнадцатиюродный",
    "двадцатиюродный",
]

_female_cousin_level = [
    "",
    "двоюродная",
    "троюродная",
    "четвероюродная",
    "пятиюродная",
    "шестиюродная",
    "семиюродная",
    "восьмиюродная",
    "девятиюродная",
    "десятиюродная",
    "одиннацатиюродная",
    "двенадцатиюродная",
    "тринадцатиюродная",
    "четырнадцатиюродная",
    "пятнадцатиюродная",
    "шестнадцатиюродная",
    "семнадцатиюродная",
    "восемнадцатиюродная",
    "девятнадцатиюродная",
    "двадцатиюродная",
]

_cousin_level = [
    "",
    "двоюродные",
    "троюродные",
    "четвероюродные",
    "пятиюродные",
    "шестиюродные",
    "семиюродные",
    "восьмиюродные",
    "девятиюродные",
    "десятиюродные",
    "одиннацатиюродные",
    "двенадцатиюродные",
    "тринадцатиюродные",
    "четырнадцатиюродные",
    "пятнадцатиюродные",
    "шестнадцатиюродные",
    "семнадцатиюродные",
    "восемнадцатиюродные",
    "девятнадцатиюродные",
    "двадцатиюродные",
]

_junior_male_removed_level = [
    "брат",
    "племянник",
    "внучатый племянник",
    "правнучатый племянник",
    "праправнучатый племянник",
    "прапраправнучатый племянник",
    "прапрапраправнучатый племянник",
]

_junior_female_removed_level = [
    "сестра",
    "племянница",
    "внучатая племянница",
    "правнучатая племянница",
    "праправнучатая племянница",
    "прапраправнучатая племянница",
    "прапрапраправнучатая племянница",
]

_juniors_removed_level = [
    "братья/сестры",
    "племянники",
    "внучатые племянники",
    "правнучатые племянники",
    "праправнучатые племянники",
    "прапраправнучатые племянники",
    "прапрапраправнучатые племянники",
]

_senior_male_removed_level = [
    "",
    "дядя",
    "дед",
    "прадед",
    "прапрадед",
    "прапрапрадед",
    "прапрапрапрадед",
]

_senior_female_removed_level = [
    "",
    "тётя",
    "бабушка",
    "прабабушка",
    "прапрабабушка",
    "прапрапрабабушка",
    "прапрапрапрабабушка",
]

_seniors_removed_level = [
    "",
    "дяди/тёти",
    "дедушки/бабушки",
    "прадеды/прабабушки",
    "прапрадеды/прапрабабушки",
    "прапрапрадеды/прапрапрабабушки",
    "прапрапрапрадеды/прапрапрапрабабушки",
]

_father_level = [
    "",
    "отец",
    "дед",
    "прадед",
    "прапрадед",
    "прапрапрадед",
    "прапрапрапрадед",
]

_mother_level = [
    "",
    "мать",
    "бабушка",
    "прабабушка",
    "прапрабабушка",
    "прапрапрабабушка",
    "прапрапрапрабабушка",
]

_son_level = [
    "",
    "сын",
    "внук",
    "правнук",
    "праправнук",
    "прапраправнук",
    "прапрапраправнук",
]

_daughter_level = [
    "",
    "дочь",
    "внучка",
    "правнучка",
    "праправнучка",
    "прапраправнучка",
    "прапрапраправнучка",
]

_children_level = [
    "",
    "дети",
    "внуки",
    "правнуки",
    "праправнуки",
    "прапраправнуки",
    "прапрапраправнуки",
    "прапрапрапраправнуки",
]

_sister_level = [
    "",
    "сестра",
    "тётя",
    "двоюродная бабушка",
    "двоюродная прабабушка",
    "двоюродная прапрабабушка",
    "двоюродная прапрапрабабушка",
    "двоюродная прапрапрапрабабушка",
]

_brother_level = [
    "",
    "брат",
    "дядя",
    "двоюродный дед",
    "двоюродный прадед",
    "двоюродный прапрадед",
    "двоюродный прапрапрадед",
    "двоюродный прапрапрапрадед",
]

_siblings_level = [
    "",
    "братья/сестры",
    "дядьки/тётки",
    "двоюродные дедушки/бабушки",
    "двоюродные прадедушки/прабабушки",
    "двоюродные прапрадедушки/прапрабабушки (5 поколение)",
    "двоюродные прапрапрадедушки/прапрапрабабушки (6 поколение)",
    "двоюродные прапрапрапрадедушки/прапрапрапрабабушки (7 поколение)",
    "двоюродные прапрапрапрапрадедушки/прапрапрапрапрабабушки (8 поколение)",
]

_nephew_level = [
    "",
    "племянник",
    "внучатый племянник",
    "правнучатый племянник",
    "праправнучатый племянник",
    "прапраправнучатый племянник",
    "прапрапраправнучатый племянник",
]

_niece_level = [
    "",
    "племянница",
    "внучатая племянница",
    "правнучатая племянница",
    "праправнучатая племянница",
    "прапраправнучатая племянница",
    "прапрапраправнучатая племянница",
]

_nephews_nieces_level = [
    "",
    "братья/сестры",
    "племянники",
    "внучатые племянники",
    "правнучатые племянники",
    "праправнучатые племянники",
    "прапраправнучатые племянники",
    "прапрапраправнучатые племянники",
]


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
class RelationshipCalculator(gramps.gen.relationship.RelationshipCalculator):
    """
    RelationshipCalculator Class
    """

    def __init__(self):
        gramps.gen.relationship.RelationshipCalculator.__init__(self)

    def get_parents(self, level):
        if level > len(_parents_level) - 1:
            return "дальние родственники"
        else:
            return _parents_level[level]

    def get_junior_male_cousin(self, level, removed):
        if (
            removed > len(_junior_male_removed_level) - 1
            or level > len(_male_cousin_level) - 1
        ):
            return "дальний родственник"
        else:
            return "%s %s" % (
                _male_cousin_level[level],
                _junior_male_removed_level[removed],
            )

    def get_senior_male_cousin(self, level, removed):
        if (
            removed > len(_senior_male_removed_level) - 1
            or level > len(_male_cousin_level) - 1
        ):
            return "дальний родственник"
        else:
            return "%s %s" % (
                _male_cousin_level[level],
                _senior_male_removed_level[removed],
            )

    def get_junior_female_cousin(self, level, removed):
        if (
            removed > len(_junior_female_removed_level) - 1
            or level > len(_male_cousin_level) - 1
        ):
            return "дальняя родственница"
        else:
            return "%s %s" % (
                _female_cousin_level[level],
                _junior_female_removed_level[removed],
            )

    def get_senior_female_cousin(self, level, removed):
        if (
            removed > len(_senior_female_removed_level) - 1
            or level > len(_male_cousin_level) - 1
        ):
            return "дальняя родственница"
        else:
            return "%s %s" % (
                _female_cousin_level[level],
                _senior_female_removed_level[removed],
            )

    def get_father(self, level):
        if level > len(_father_level) - 1:
            return "дальний предок"
        else:
            return _father_level[level]

    def get_son(self, level):
        if level > len(_son_level) - 1:
            return "дальний потомок"
        else:
            return _son_level[level]

    def get_mother(self, level):
        if level > len(_mother_level) - 1:
            return "дальний предок"
        else:
            return _mother_level[level]

    def get_daughter(self, level):
        if level > len(_daughter_level) - 1:
            return "дальний потомок"
        else:
            return _daughter_level[level]

    def _get_aunt(self, level, step="", inlaw=""):
        if level > len(_sister_level) - 1:
            return "дальний предок в соседнем поколении"
        else:
            return _sister_level[level]

    def _get_uncle(self, level, step="", inlaw=""):
        if level > len(_brother_level) - 1:
            return "дальний предок в соседнем поколении"
        else:
            return _brother_level[level]

    def _get_sibling(self, level, step="", inlaw=""):
        """
        Sibling of unknown gender
        """
        return (
            self._get_uncle(level, step, inlaw)
            + " или "
            + self._get_aunt(level, step, inlaw)
        )

    def get_nephew(self, level):
        if level > len(_nephew_level) - 1:
            return "дальний потомок в соседнем поколении"
        else:
            return _nephew_level[level]

    def get_niece(self, level):
        if level > len(_niece_level) - 1:
            return "дальний потомок в соседнем поколении"
        else:
            return _niece_level[level]

    def get_single_relationship_string(
        self,
        Ga,
        Gb,
        gender_a,
        gender_b,
        reltocommon_a,
        reltocommon_b,
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        if Gb == 0:
            if Ga == 0:
                return "один человек"
            elif gender_b == Person.MALE:
                return self.get_father(Ga)
            else:
                return self.get_mother(Ga)
        elif Ga == 0:
            if gender_b == Person.MALE:
                return self.get_son(Gb)
            else:
                return self.get_daughter(Gb)
        elif Gb == 1:
            if gender_b == Person.MALE:
                return self._get_uncle(Ga)
            else:
                return self._get_aunt(Ga)
        elif Ga == 1:
            if gender_b == Person.MALE:
                return self.get_nephew(Gb - 1)
            else:
                return self.get_niece(Gb - 1)
        elif Ga > Gb:
            if gender_b == Person.MALE:
                return self.get_senior_male_cousin(Gb - 1, Ga - Gb)
            else:
                return self.get_senior_female_cousin(Gb - 1, Ga - Gb)
        else:
            if gender_b == Person.MALE:
                return self.get_junior_male_cousin(Ga - 1, Gb - Ga)
            else:
                return self.get_junior_female_cousin(Ga - 1, Gb - Ga)

    def get_plural_relationship_string(
        self,
        Ga,
        Gb,
        reltocommon_a="",
        reltocommon_b="",
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        rel_str = "дальние родственники"
        if Ga == 0:
            # These are descendants
            if Gb < len(_children_level):
                rel_str = _children_level[Gb]
            else:
                rel_str = "дальние потомки"
        elif Gb == 0:
            # These are parents/grand parents
            if Ga < len(_parents_level):
                rel_str = _parents_level[Ga]
            else:
                rel_str = "дальние предки"
        elif Gb == 1:
            # These are siblings/aunts/uncles
            if Ga < len(_siblings_level):
                rel_str = _siblings_level[Ga]
            else:
                rel_str = "дальние дяди/тёти"
        elif Ga == 1:
            # These are nieces/nephews
            if Gb < len(_nephews_nieces_level):
                rel_str = _nephews_nieces_level[Gb]
            else:
                rel_str = "дальние племянники/племянницы"
        elif Ga > Gb:
            # These are cousins in different generations with the second person
            # being in a higher generation from the common ancestor than the
            # first person.
            if Gb <= len(_seniors_removed_level) and (Ga - Gb) < len(_cousin_level):
                rel_str = "%s %s" % (
                    _cousin_level[Gb - 1],
                    _seniors_removed_level[Ga - Gb],
                )
            else:
                rel_str = "(старшие) дальние родственники"
        else:
            # These are cousins in different generations with the second person
            # being in a lower generation from the common ancestor than the
            # first person.
            if Ga <= len(_juniors_removed_level) and (Gb - Ga) < len(_cousin_level):
                rel_str = "%s %s" % (
                    _cousin_level[Ga - 1],
                    _juniors_removed_level[Gb - Ga],
                )
            else:
                rel_str = "(младшие) дальние родственники"

        if in_law_b == True:
            # TODO: Translate this!
            rel_str = "spouses of %s" % rel_str

        return rel_str


# TODO: def get_sibling_relationship_string for Russian step and inlaw relations

if __name__ == "__main__":
    # Test function. Call it as follows from the command line (so as to find
    #        imported modules):
    #    export PYTHONPATH=/path/to/gramps/src
    # python src/plugins/rel/rel_ru.py
    # (Above not needed here)

    """TRANSLATORS, copy this if statement at the bottom of your
    rel_xx.py module, and test your work with:
    python src/plugins/rel/rel_xx.py
    """
    from gramps.gen.relationship import test

    RC = RelationshipCalculator()
    test(RC, True)
